#include <math.h>
#include <time.h>
#include <stdio.h>
#include <ctype.h>
#include <string.h>
#include <stdlib.h>
#include <stdarg.h>
#include <malloc.h>
#include "3dtools.h"
#include "3dadds.h"
#include "3dcalc.h"
#include "3dmisc.h"
#include "3dfilter.h"
#include "Multi_IO.h"


/************************************************************************
[NAME]
CSrebin3D\_TL

[SYNOSPSIS]
MultiStruct * CSrebin3D_TL(MultiStruct*,TrigDef*,XYZDef*);

[DESCRIPTION] 
Calculates the two-dimensional Fourier transformation of a sinogram 
(for each combination of angles). This sinogram is distributed in
the Fourier space of the reconstructed volume with tri linear interpolation. 
(Central Slice ifea).

[USAGE]
{\tt RecVol=CSrebin3D\_Tl(Sinogram,MyTrig,MyXYZ);}

Reconstructs the volume from the sinogram using central slice theorem
for lines in a three dimensional space.

[REVISION]
Nov 14 Ptoft\\
10/12-95 Peter Toft - Dropped OrthoVecors and made 
use of individual vectors. See TrigDef.\\
13/12-95 Peter Toft. Error correction. \\
14/12-95 Peter Toft. Experimental !!
***************************************************************************/
MultiStruct * CSrebin3D_Tl(MultiStruct *MySino,
                           TrigDef *MyTrig,
                           XYZDef *MyXYZ)
{
  int adr,adr3D,adr3Dz,adr3Dyz,adr3Dxyz,adrd3D;
  int adrspec,ThetaPhiUSamples,off,thetaint,phiint,uuint,vvint;
  int xint,yint,zint;
  int XSamples,YSamples,ZSamples,Pow2XSamples,Pow2YSamples,Pow2ZSamples;
  int ThetaSamples,PhiSamples,USamples,VSamples,Pow2USamples,Pow2VSamples;
  int XSampleshm1,YSampleshm1,ZSampleshm1;
  int nu_xint,nu_yint,nu_zint,nu_xint2,nu_yint2,nu_zint2;
  int trigoff,ThetaPhiSamples;
  int *XYZSamples,*Pow2XYZSamples,*Pow2UVSamples;
  float cstx,csty,cstz,weight,dx,dy,dz;
  float nu_xint_f,nu_yint_f,nu_zint_f;
  float *spectrum;
  float alpha_x,alpha_y,alpha_z;
  float beta_x,beta_y,beta_z;
  float nu_x,nu_y,nu_z;
  float DeltaU,DeltaV,DeltaX,DeltaY,DeltaZ;
  float inv_nu_u,inv_nu_v;
  float nu_u,nu_v;
  float reldx,reldy,reldz;
  MultiStruct *WeightVol,*ReconVol,*ComplexRecon;
  float *ar_alpha_x,*ar_alpha_y,*ar_alpha_z;
  float *ar_beta_x,*ar_beta_y,*ar_beta_z;
  float *nu_u_ar,*nu_v_ar;
  float *SinoSignal,*ComplexSignal,*WeightSignal,*ReconSignal;
  int *shift_u,*shift_v,off2;
  int USampleshm1,VSampleshm1,dPow2USamples;
  float oneminusreldx,oneminusreldy,oneminusreldz;
  float nualpx,nualpy,nualpz,realsignal,imagsignal;
  float deltafrac;

  Print(_DNormal,"(CSrebin3D_Tl)\n");
  if (MySino->Dim!=4)
    Error("Wrong dimension of MultiStruct (CSrebin3D_Tl)");
  if (MySino->ArrayType!=_RealArray)
    Error("Wrong Array Type. Must be RealArray (CSrebin3D_Tl)");
              
  Pow2UVSamples=IntVector(2);
  Pow2XYZSamples=IntVector(3);
  XYZSamples=IntVector(3);

  ThetaSamples=MySino->Samples[0];
  PhiSamples=MySino->Samples[1];
  USamples=MySino->Samples[2];
  VSamples=MySino->Samples[3];
  Pow2USamples=NextRadix2(USamples);
  Pow2VSamples=NextRadix2(VSamples);
  XSamples=MyXYZ->XSamples;
  YSamples=MyXYZ->YSamples;
  ZSamples=MyXYZ->ZSamples;

  XSampleshm1=(XSamples-1)/2;
  YSampleshm1=(YSamples-1)/2;
  ZSampleshm1=(ZSamples-1)/2;

  Pow2UVSamples[0]=Pow2USamples;
  Pow2UVSamples[1]=Pow2VSamples;
  XYZSamples[0]=XSamples;
  XYZSamples[1]=YSamples;
  XYZSamples[2]=ZSamples;

  shift_u=IntVector(USamples);
  shift_v=IntVector(VSamples);

  DeltaU=MyTrig->DeltaU;
  DeltaV=MyTrig->DeltaV;

  Pow2XSamples=Pow2XYZSamples[0]=NextRadix2(XYZSamples[0]);
  Pow2YSamples=Pow2XYZSamples[1]=NextRadix2(XYZSamples[1]);
  Pow2ZSamples=Pow2XYZSamples[2]=NextRadix2(XYZSamples[2]);
  ReconVol=NewMultiStruct("Recon",3,XYZSamples,_RealArray);
  WeightVol=NewMultiStruct("Weight",3,Pow2XYZSamples,_RealArray);
  ComplexRecon=NewMultiStruct("Spectrum3D",3,Pow2XYZSamples,_ComplexArray);
  ComplexSignal=ComplexRecon->Signal;

  ReconVol->Delta[0]=DeltaX=MyXYZ->DeltaX;
  ReconVol->Delta[1]=DeltaY=MyXYZ->DeltaY;
  ReconVol->Delta[2]=DeltaZ=MyXYZ->DeltaZ;

  ReconVol->Min[0]=MyXYZ->Xmin;
  ReconVol->Min[1]=MyXYZ->Ymin;
  ReconVol->Min[2]=MyXYZ->Zmin;

  deltafrac=DeltaU*DeltaV/(DeltaX*DeltaY*DeltaZ);

  ar_alpha_x=MyTrig->alpha_x;
  ar_alpha_y=MyTrig->alpha_y;
  ar_alpha_z=MyTrig->alpha_z;
  ar_beta_x=MyTrig->beta_x;
  ar_beta_y=MyTrig->beta_y;
  ar_beta_z=MyTrig->beta_z;

  SinoSignal=MySino->Signal;
  WeightSignal=WeightVol->Signal;
  ReconSignal=ReconVol->Signal;

  inv_nu_u=1.0/(Pow2USamples*DeltaU);
  inv_nu_v=1.0/(Pow2VSamples*DeltaV);

  /* Half of the spacial sampling frequency in each direction */
  cstx=0.5/DeltaX;
  csty=0.5/DeltaY;
  cstz=0.5/DeltaZ;

  ThetaPhiSamples=ThetaSamples*PhiSamples;
  spectrum=FloatVector(Pow2USamples*Pow2VSamples*2);

  ThetaPhiUSamples=ThetaSamples*PhiSamples*USamples;  
  dPow2USamples=2*Pow2USamples;
  nu_u_ar=FloatVector(Pow2USamples);
  nu_v_ar=FloatVector(Pow2VSamples);

  /* shift_u[i] means that this sample should go to shift_u[i] 
     when doing FFT                                            */
  USampleshm1=(USamples-1)/2;
  for (uuint=0;uuint<USampleshm1;uuint++)
    shift_u[uuint]=uuint+Pow2USamples-USampleshm1; 
  for (uuint=USampleshm1;uuint<USamples;uuint++)
    shift_u[uuint]=uuint-USampleshm1;

  /* Same procedure for the v direction */
  VSampleshm1=(VSamples-1)/2;
  for (vvint=0;vvint<VSampleshm1;vvint++)
    shift_v[vvint]=vvint+Pow2VSamples-VSampleshm1; 
  for (vvint=VSampleshm1;vvint<VSamples;vvint++)
    shift_v[vvint]=vvint-VSampleshm1;

  /* Initialization for the frequency/sample relation */
  for (uuint=0;uuint<Pow2USamples/2;uuint++)
    nu_u_ar[uuint]=uuint*inv_nu_u;
  for (uuint=Pow2USamples/2+1;uuint<Pow2USamples;uuint++)
    nu_u_ar[uuint]=(uuint-Pow2USamples)*inv_nu_u;

  /* Same for the v-direction */
  for (vvint=0;vvint<Pow2VSamples;vvint++)
    nu_v_ar[vvint]=vvint*inv_nu_v;
  for (vvint=Pow2VSamples/2+1;vvint<Pow2VSamples;vvint++)
    nu_v_ar[vvint]=(vvint-Pow2VSamples)*inv_nu_v;

  trigoff=-1;
  for (phiint=0;phiint<PhiSamples;phiint++)
    for (thetaint=0;thetaint<ThetaSamples;thetaint++)
    {
      trigoff++;
      alpha_x=ar_alpha_x[trigoff];
      alpha_y=ar_alpha_y[trigoff];
      alpha_z=ar_alpha_z[trigoff];
      
      beta_x=ar_beta_x[trigoff];
      beta_y=ar_beta_y[trigoff];
      beta_z=ar_beta_z[trigoff];
      ZeroFloatVector(spectrum,Pow2USamples*2*Pow2USamples);
      for (uuint=0;uuint<USamples;uuint++)
      {
        off=trigoff+ThetaPhiSamples*uuint;
        off2=2*shift_u[uuint];
        for (vvint=0;vvint<VSamples;vvint++)
        {
          adr=off2+shift_v[vvint]*dPow2USamples;
          spectrum[adr++]=SinoSignal[off+ThetaPhiUSamples*vvint];
          spectrum[adr]=0.0;
        }
      }
      for (uuint=USampleshm1+1;uuint<Pow2USamples-USampleshm1;uuint++)
        for (vvint=0;vvint<Pow2VSamples;vvint++)
        {
          off2=2*(uuint+vvint*Pow2USamples);
          spectrum[off2]=0.0;
          spectrum[off2+1]=0.0;
        }
      for (vvint=VSampleshm1+1;vvint<Pow2VSamples-VSampleshm1;vvint++)
        for (uuint=0;uuint<Pow2USamples;uuint++)
        {
          off2=2*(uuint+vvint*Pow2USamples);
          spectrum[off2]=0.0;
          spectrum[off2+1]=0.0;
        }

      /* Now data is reoorganized and the 2D FFT is performed */
      NdimFFT(spectrum,Pow2UVSamples,2,_FFT);

      /* After the 2D FFT is performed the data 
         is distributed in the 3D Fourier space  */

      for (uuint=0;uuint<Pow2USamples;uuint++)
      {
        nu_u=nu_u_ar[uuint];

        nualpx=alpha_x*nu_u;
        nualpy=alpha_y*nu_u;
        nualpz=alpha_z*nu_u;

        for (vvint=0;vvint<Pow2VSamples;vvint++)
        {
          nu_v=nu_v_ar[vvint];

          nu_x=nualpx+beta_x*nu_v;
          nu_y=nualpy+beta_y*nu_v;
          nu_z=nualpz+beta_z*nu_v;
          
          if ((nu_x<cstx) && (nu_x>-cstx) 
              && (nu_y<csty) && (nu_y>-csty) 
              && (nu_z<cstz) && (nu_z>-cstz))
          {
            nu_xint_f=nu_x*DeltaX*Pow2XSamples;
            nu_xint=floor(nu_xint_f);
            nu_xint2=nu_xint+1;
            dx=nu_xint_f-nu_xint;
            reldx=dx/DeltaX;
            oneminusreldx=1.0-reldx;
            if (nu_xint<0) nu_xint+=Pow2XSamples;
            if (nu_xint2<0) nu_xint2+=Pow2XSamples;

            nu_yint_f=nu_y*DeltaY*Pow2YSamples;
            nu_yint=floor(nu_yint_f);
            nu_yint2=nu_yint+1;
            dy=nu_yint_f-nu_yint;
            reldy=dy/DeltaY;
            oneminusreldy=1.0-reldy;
            if (nu_yint<0) nu_yint+=Pow2YSamples;
            if (nu_yint2<0) nu_yint2+=Pow2YSamples;
            
            nu_zint_f=nu_z*DeltaZ*Pow2ZSamples;
            nu_zint=floor(nu_zint_f);
            nu_zint2=nu_zint+1;
            dz=nu_zint_f-nu_zint;
            reldz=dz/DeltaZ;
            oneminusreldz=1.0-reldz;
            if (nu_zint<0) nu_zint+=Pow2ZSamples;
            if (nu_zint2<0) nu_zint2+=Pow2ZSamples;
            
            /* Tri linear interpolation */
            adrspec=2*(uuint+vvint*Pow2USamples);
            realsignal=spectrum[adrspec];
            imagsignal=spectrum[adrspec+1];
            
            /*x:low y:low z:low*/
            adr3D=nu_xint+Pow2XSamples*(nu_yint+Pow2YSamples*nu_zint);
            adrd3D=adr3D*2;
            weight=oneminusreldx*oneminusreldy*oneminusreldz;
            WeightSignal[adr3D]+=weight;
            ComplexSignal[adrd3D++]+=weight*realsignal;
            ComplexSignal[adrd3D]+=weight*imagsignal;
            
            /*x:high y:low z:low*/
            adr3D=nu_xint2+Pow2XSamples*(nu_yint+Pow2YSamples*nu_zint);
            adrd3D=adr3D*2;
            weight=reldx*oneminusreldy*oneminusreldz;
            WeightSignal[adr3D]+=weight;
            ComplexSignal[adrd3D++]+=weight*realsignal;
            ComplexSignal[adrd3D]+=weight*imagsignal;
            
            /*x:low y:high z:low*/
            adr3D=nu_xint+Pow2XSamples*(nu_yint2+Pow2YSamples*nu_zint);
            adrd3D=adr3D*2;
            weight=oneminusreldx*reldy*oneminusreldz;
            WeightSignal[adr3D]+=weight;
            ComplexSignal[adrd3D++]+=weight*realsignal;
            ComplexSignal[adrd3D]+=weight*imagsignal;
            
            /*x:high y:high z:low*/
            adr3D=nu_xint2+Pow2XSamples*(nu_yint2+Pow2YSamples*nu_zint);
            adrd3D=adr3D*2;
            weight=reldx*reldy*oneminusreldz;
            WeightSignal[adr3D]+=weight;
            ComplexSignal[adrd3D++]+=weight*realsignal;
            ComplexSignal[adrd3D]+=weight*imagsignal;
            
            /*x:low y:low z:high*/
            adr3D=nu_xint+Pow2XSamples*(nu_yint+Pow2YSamples*nu_zint2);
            adrd3D=adr3D*2;
            weight=oneminusreldx*oneminusreldy*reldz;
            WeightSignal[adr3D]+=weight;
            ComplexSignal[adrd3D++]+=weight*realsignal;
            ComplexSignal[adrd3D]+=weight*imagsignal;
            
            /*x:high y:low z:high*/
            adr3D=nu_xint2+Pow2XSamples*(nu_yint+Pow2YSamples*nu_zint2);
            adrd3D=adr3D*2;
            weight=reldx*oneminusreldy*reldz;
            WeightSignal[adr3D]+=weight;
            ComplexSignal[adrd3D++]+=weight*realsignal;
            ComplexSignal[adrd3D]+=weight*imagsignal;

            /*x:low y:high z:high*/
            adr3D=nu_xint+Pow2XSamples*(nu_yint2+Pow2YSamples*nu_zint2);;
            adrd3D=adr3D*2;
            weight=oneminusreldx*reldy*reldz;
            WeightSignal[adr3D]+=weight;
            ComplexSignal[adrd3D++]+=weight*realsignal;
            ComplexSignal[adrd3D]+=weight*imagsignal;
            
            /*x:high y:high z:high*/
            adr3D=nu_xint2+Pow2XSamples*(nu_yint2+Pow2YSamples*nu_zint2);
            adrd3D=adr3D*2;
            weight=reldx*reldy*reldz;
            WeightSignal[adr3D]+=weight;
            ComplexSignal[adrd3D++]+=weight*realsignal;
            ComplexSignal[adrd3D]+=weight*imagsignal;
          }
        }
      }
    }

  /* Normalization using WeightVol */
  
  adr3D=0;
  adrd3D=0;
  for (adr=0;adr<Pow2XSamples*Pow2YSamples*Pow2ZSamples;adr++)
  {
    weight=WeightSignal[adr3D++];
    if (weight>0.1)
    {
      ComplexSignal[adrd3D++]/=weight;
      ComplexSignal[adrd3D++]/=weight;
     }
  }

  /* Inverse 3D FFT to recover volume data */
  NdimFFT(ComplexSignal,Pow2XYZSamples,3,_IFFT);

  adr=0;
  for (zint=0;zint<ZSamples;zint++)
  {
    if (zint>=ZSampleshm1)
      adr3Dz=zint-ZSampleshm1;
    else
      adr3Dz=Pow2ZSamples-(ZSampleshm1-zint);
    for (yint=0;yint<YSamples;yint++)
    {
      if (yint>=YSampleshm1)
        adr3Dyz=yint-YSampleshm1+adr3Dz*YSamples;
      else
        adr3Dyz=Pow2YSamples-(YSamples-yint)+adr3Dz*Pow2YSamples;
      for (xint=0;xint<XSamples;xint++)
      {
        if (xint>=XSampleshm1)
          adr3Dxyz=xint-XSampleshm1+adr3Dyz*XSamples;
        else
          adr3Dxyz=Pow2XSamples-(XSamples-xint)+adr3Dyz*Pow2XSamples;
        ReconSignal[adr++]=ComplexSignal[2*adr3Dxyz];
      }
    }
  }
  FreeMultiStruct(WeightVol);
  FreeMultiStruct(ComplexRecon);
  Free(spectrum);
  Free(nu_u_ar);
  Free(nu_v_ar);
  Free(shift_u);
  Free(shift_v);
  return ReconVol;
}


/************************************************************************
[NAME]
CSrebin3D\_NN

[SYNOSPSIS]
 MultiStruct * CSrebin3D\_NN(MultiStruct*,TrigDef*,XYZDef*);

[DESCRIPTION] 
Calculates the two-dimensional Fourier transformation of a sinogram 
(for each combination of angles). This sinogram is distributed in
the Fourier space of the reconstructed volume. (Central Slice ifea).

[USAGE]
{\tt RecVol=CSrebin3D\_NN(Sinogram,MyTrig,MyXYZ);}

Reconstructs the volume from the sinogram using central slice theorem
for lines in a three dimensional space.

[REVISION]
Nov 14 Ptoft\\
10/12-95 Peter Toft - Dropped OrthoVecors and made 
use of individual vectors. See TrigDef.\\
13/12-95 Peter Toft. Error correction. \\
14/12-95 Peter Toft. Errors - still experimental
***************************************************************************/
MultiStruct * CSrebin3D_NN(MultiStruct *MySino,
                           TrigDef *MyTrig,
                           XYZDef *MyXYZ)
{
  int adr,adr3D,adr3Dz,adr3Dyz,adr3Dxyz,adrd3D;
  int adrspec,ThetaPhiUSamples,off,thetaint,phiint,uuint,vvint;
  int xint,yint,zint;
  int XSamples,YSamples,ZSamples,Pow2XSamples,Pow2YSamples,Pow2ZSamples;
  int ThetaSamples,PhiSamples,USamples,VSamples,Pow2USamples,Pow2VSamples;
  int XSampleshm1,YSampleshm1,ZSampleshm1;
  int nu_xint,nu_yint,nu_zint;
  int trigoff,ThetaPhiSamples;
  int *XYZSamples,*Pow2XYZSamples,*Pow2UVSamples;
  float cstx,csty,cstz,weight;
  float nu_xint_f,nu_yint_f,nu_zint_f;
  float *spectrum;
  float alpha_x,alpha_y,alpha_z;
  float beta_x,beta_y,beta_z;
  float nu_x,nu_y,nu_z;
  float deltafrac;
  float DeltaU,DeltaV,DeltaX,DeltaY,DeltaZ;
  float inv_nu_u,inv_nu_v;
  float nu_u,nu_v;
  MultiStruct *WeightVol,*ReconVol,*ComplexRecon;
  float *ar_alpha_x,*ar_alpha_y,*ar_alpha_z;
  float *ar_beta_x,*ar_beta_y,*ar_beta_z;
  float *nu_u_ar,*nu_v_ar;
  float *SinoSignal,*ComplexSignal,*WeightSignal,*ReconSignal;
  int *shift_u,*shift_v,off2;
  int USampleshm1,VSampleshm1,dPow2USamples;
  float nualpx,nualpy,nualpz,realsignal,imagsignal;
  MultiStruct *ged;

  if (MySino->Dim!=4)
    Error("Wrong dimension of MultiStruct (CSrebin3D)");
  if (MySino->ArrayType!=_RealArray)
    Error("Wrong Array Type. Must be RealArray (CSrebin3D)");
              
  Pow2UVSamples=IntVector(2);
  Pow2XYZSamples=IntVector(3);
  XYZSamples=IntVector(3);

  ThetaSamples=MySino->Samples[0];
  PhiSamples=MySino->Samples[1];
  USamples=MySino->Samples[2];
  VSamples=MySino->Samples[3];
  Pow2USamples=NextRadix2(USamples);
  Pow2VSamples=NextRadix2(VSamples);
  XSamples=MyXYZ->XSamples;
  YSamples=MyXYZ->YSamples;
  ZSamples=MyXYZ->ZSamples;

  XSampleshm1=(XSamples-1)/2;
  YSampleshm1=(YSamples-1)/2;
  ZSampleshm1=(ZSamples-1)/2;

  Pow2UVSamples[0]=Pow2USamples;
  Pow2UVSamples[1]=Pow2VSamples;
  XYZSamples[0]=XSamples;
  XYZSamples[1]=YSamples;
  XYZSamples[2]=ZSamples;

  shift_u=IntVector(USamples);
  shift_v=IntVector(VSamples);

  Pow2XSamples=Pow2XYZSamples[0]=NextRadix2(XYZSamples[0]);
  Pow2YSamples=Pow2XYZSamples[1]=NextRadix2(XYZSamples[1]);
  Pow2ZSamples=Pow2XYZSamples[2]=NextRadix2(XYZSamples[2]);
  ReconVol=NewMultiStruct("Recon",3,XYZSamples,_RealArray);
  WeightVol=NewMultiStruct("Weight",3,Pow2XYZSamples,_RealArray);
  ComplexRecon=NewMultiStruct("Spectrum3D",3,Pow2XYZSamples,_ComplexArray);
  ComplexSignal=ComplexRecon->Signal;

  DeltaU=MyTrig->DeltaU;
  DeltaV=MyTrig->DeltaV;

  ReconVol->Delta[0]=DeltaX=MyXYZ->DeltaX;
  ReconVol->Delta[1]=DeltaY=MyXYZ->DeltaY;
  ReconVol->Delta[2]=DeltaZ=MyXYZ->DeltaZ;

  ReconVol->Min[0]=MyXYZ->Xmin;
  ReconVol->Min[1]=MyXYZ->Ymin;
  ReconVol->Min[2]=MyXYZ->Zmin;

  deltafrac=DeltaU*DeltaV/(DeltaX*DeltaY*DeltaZ);

  ar_alpha_x=MyTrig->alpha_x;
  ar_alpha_y=MyTrig->alpha_y;
  ar_alpha_z=MyTrig->alpha_z;
  ar_beta_x=MyTrig->beta_x;
  ar_beta_y=MyTrig->beta_y;
  ar_beta_z=MyTrig->beta_z;

  SinoSignal=MySino->Signal;
  WeightSignal=WeightVol->Signal;
  ReconSignal=ReconVol->Signal;

  inv_nu_u=1.0/(Pow2USamples*DeltaU);
  inv_nu_v=1.0/(Pow2VSamples*DeltaV);

  /* Half of the spacial sampling frequency in each direction */
  cstx=0.5/DeltaX;
  csty=0.5/DeltaY;
  cstz=0.5/DeltaZ;

  ThetaPhiSamples=ThetaSamples*PhiSamples;
  spectrum=FloatVector(Pow2USamples*Pow2VSamples*2);

  ThetaPhiUSamples=ThetaSamples*PhiSamples*USamples;  
  dPow2USamples=2*Pow2USamples;
  nu_u_ar=FloatVector(Pow2USamples);
  nu_v_ar=FloatVector(Pow2VSamples);

  /* shift_u[i] means that this sample should 
     go to shift_u[i] when doing FFT          */
  USampleshm1=(USamples-1)/2;
  for (uuint=0;uuint<USampleshm1;uuint++)
    shift_u[uuint]=uuint+Pow2USamples-USampleshm1; 
  for (uuint=USampleshm1;uuint<USamples;uuint++)
    shift_u[uuint]=uuint-USampleshm1;

  /* Same procedure for the v direction */
  VSampleshm1=(VSamples-1)/2;
  for (vvint=0;vvint<VSampleshm1;vvint++)
    shift_v[vvint]=vvint+Pow2VSamples-VSampleshm1; 
  for (vvint=VSampleshm1;vvint<VSamples;vvint++)
    shift_v[vvint]=vvint-VSampleshm1;

  /* Initialization for the frequency/sample relation */
  for (uuint=0;uuint<Pow2USamples/2;uuint++)
    nu_u_ar[uuint]=uuint*inv_nu_u;
  for (uuint=Pow2USamples/2+1;uuint<Pow2USamples;uuint++)
    nu_u_ar[uuint]=(uuint-Pow2USamples)*inv_nu_u;

  /* Same for the v-direction */
  for (vvint=0;vvint<Pow2VSamples;vvint++)
    nu_v_ar[vvint]=vvint*inv_nu_v;
  for (vvint=Pow2VSamples/2+1;vvint<Pow2VSamples;vvint++)
    nu_v_ar[vvint]=(vvint-Pow2VSamples)*inv_nu_v;

  Print(_DNormal,"Forward FFT of Sinogram and mapping into 3D Fourier space\n");
  trigoff=-1;
  for (phiint=0;phiint<PhiSamples;phiint++)
    for (thetaint=0;thetaint<ThetaSamples;thetaint++)
    {
      trigoff++;
      alpha_x=ar_alpha_x[trigoff];
      alpha_y=ar_alpha_y[trigoff];
      alpha_z=ar_alpha_z[trigoff];
      
      beta_x=ar_beta_x[trigoff];
      beta_y=ar_beta_y[trigoff];
      beta_z=ar_beta_z[trigoff];

      for (uuint=0;uuint<USamples;uuint++)
      {
        off=trigoff+ThetaPhiSamples*uuint;
        off2=2*shift_u[uuint];
        for (vvint=0;vvint<VSamples;vvint++)
        {
          adr=off2+shift_v[vvint]*dPow2USamples;
          spectrum[adr++]=SinoSignal[off+ThetaPhiUSamples*vvint];
          spectrum[adr]=0.0;
        }
      }
      for (uuint=USampleshm1+1;uuint<Pow2USamples-USampleshm1;uuint++)
        for (vvint=0;vvint<Pow2VSamples;vvint++)
        {
          off2=2*(uuint+vvint*Pow2USamples);
          spectrum[off2]=0.0;
          spectrum[off2+1]=0.0;
        }
      for (vvint=VSampleshm1+1;vvint<Pow2VSamples-VSampleshm1;vvint++)
        for (uuint=0;uuint<Pow2USamples;uuint++)
        {
          off2=2*(uuint+vvint*Pow2USamples);
          spectrum[off2]=0.0;
          spectrum[off2+1]=0.0;
        }


      /* Now data is reoorganized and the 2D FFT is performed */
      NdimFFT(spectrum,Pow2UVSamples,2,_FFT);

      /* After the 2D FFT is performed the data 
         is distributed in the 3D Fourier space  */
      for (uuint=0;uuint<Pow2USamples;uuint++)
      {
        nu_u=nu_u_ar[uuint];

        nualpx=alpha_x*nu_u;
        nualpy=alpha_y*nu_u;
        nualpz=alpha_z*nu_u;

        for (vvint=0;vvint<Pow2VSamples;vvint++)
        {
          nu_v=nu_v_ar[vvint];

          nu_x=nualpx+beta_x*nu_v;
          nu_y=nualpy+beta_y*nu_v;
          nu_z=nualpz+beta_z*nu_v;
          
          if ((nu_x<cstx) && (nu_x>-cstx) 
              && (nu_y<csty) && (nu_y>-csty) 
              && (nu_z<cstz) && (nu_z>-cstz))
          {
            nu_xint_f=nu_x*DeltaX*Pow2XSamples;
            nu_xint=floor(nu_xint_f+0.5);
            if (nu_xint<0) nu_xint+=Pow2XSamples;

            nu_yint_f=nu_y*DeltaY*Pow2YSamples;
            nu_yint=floor(nu_yint_f+0.5);
            if (nu_yint<0) nu_yint+=Pow2YSamples;
            
            nu_zint_f=nu_z*DeltaZ*Pow2ZSamples;
            nu_zint=floor(nu_zint_f+0.5);
            if (nu_zint<0) nu_zint+=Pow2ZSamples;
            
            adrspec=2*(uuint+vvint*Pow2USamples);
            realsignal=spectrum[adrspec];
            imagsignal=spectrum[adrspec+1];
            
            /*x:low y:low z:low*/
            adr3D=nu_xint+Pow2XSamples*(nu_yint+Pow2YSamples*nu_zint);
            adrd3D=adr3D*2;
            WeightSignal[adr3D]+=1;
            ComplexSignal[adrd3D++]+=realsignal;
            ComplexSignal[adrd3D]+=imagsignal;
          }
        }
      }
    }

  Print(_DNormal,"Normalization\n");

  /* Normalization using WeightVol */
  
  adr3D=0;
  adrd3D=0;
  for (adr=0;adr<Pow2XSamples*Pow2YSamples*Pow2ZSamples;adr++)
  {
    weight=WeightSignal[adr3D++];
    if (weight>0)
    {
      ComplexSignal[adrd3D++]/=weight;
      ComplexSignal[adrd3D++]/=weight;
     }
  }

  Print(_DNormal,"Inverse 3D FFT of spectrum\n");

  ged=DuplicateMultiStruct(ComplexRecon);
  ImagMultiStruct(ged);
  RenameMultiStruct(ged,"Imag");
  WriteMultiStruct(ged);

  /* Inverse 3D FFT to recover volume data */
  NdimFFT(ComplexSignal,Pow2XYZSamples,3,_IFFT);

  Print(_DNormal,"Shifting volume data back\n");
  adr=0;
  for (zint=0;zint<ZSamples;zint++)
  {
    if (zint>=ZSampleshm1)
      adr3Dz=zint-ZSampleshm1;
    else
      adr3Dz=Pow2ZSamples-(ZSampleshm1-zint);
    for (yint=0;yint<YSamples;yint++)
    {
      if (yint>=YSampleshm1)
        adr3Dyz=yint-YSampleshm1+adr3Dz*YSamples;
      else
        adr3Dyz=Pow2YSamples-(YSamples-yint)+adr3Dz*Pow2YSamples;
      for (xint=0;xint<XSamples;xint++)
      {
        if (xint>=XSampleshm1)
          adr3Dxyz=xint-XSampleshm1+adr3Dyz*XSamples;
        else
          adr3Dxyz=Pow2XSamples-(XSamples-xint)+adr3Dyz*Pow2XSamples;
        ReconSignal[adr++]=ComplexSignal[2*adr3Dxyz]*deltafrac;
      }
    }
  }
  FreeMultiStruct(WeightVol);
  FreeMultiStruct(ComplexRecon);
  Free(spectrum);
  Free(nu_u_ar);
  Free(nu_v_ar);
  Free(shift_u);
  Free(shift_v);
  return ReconVol;
}



/****************************************************************
[NAME]
NumericalRadon3D\_Restore

[SYNOPSIS]
 MultiStruct* NumericalRadon3D_Restore(MultiStruct *MyRadon,
                                       MultiStruct *InVolume,
                                       TrigDef *MyTrig)

[DESCRIPTION]
The function will numerically calculate line integrals through
a three dimensional volume. The sampling of the Radon domain is given 
through {\tt MyTrig}.

[USAGE]
{\tt NumericalRadon3D(MyRadon,InVolume,MyTrig);}

[NOTE]
Uses bilinear interpolation. Requires that the Sinogram ({\tt MyRadon})
is allocated. See also NumericalRadon3D.

[REVISION]
20/5-95 Peter Toft\\
19/6-95 Peter Toft. (Uses rectau. Bug fixes).\\
30/6-95 Peter Toft. (Bug fixes. Rectau -> Rectau*DeltaX and so on).\\
24/8-95 Peter Toft. Split function in two parts.\\
10/12-95 Peter Toft - Dropped OrthoVecors and made 
use of individual vecvors. See TrigDef.
****************************************************************/
void NumericalRadon3D_Restore(MultiStruct *My4D,
                                      MultiStruct *InVolume,
                                      TrigDef *MyTrig)
{
  float rad,broek,u,v;
  float tau_x,tau_y,tau_z;
  float atau_x,atau_y,atau_z;
  float ax,bx,ay,by,az,bz;
  float alpha_x,alpha_y,alpha_z;
  float xsint,ysint,zsint;
  float beta_x,beta_y,beta_z;
  float DeltaX,DeltaY,DeltaZ;
  int XSamples,YSamples,ZSamples;
  int xint,yint,zint,trigoff;
  float Xmin,Ymin,Zmin;
  float rectau;
  float *sig,*outsig;
  int *Samples;
  int uuint,vvint;
  int ThetaSamples,PhiSamples,USamples,VSamples;
  float *ar_alpha_x,*ar_alpha_y,*ar_alpha_z;
  float *ar_beta_x,*ar_beta_y,*ar_beta_z;
  float *ar_tau_x,*ar_tau_y,*ar_tau_z;
  
  Print(_DDebug,"Initializing (NumericalRadon3D)\n");

  Samples=IntVector(4);
  ThetaSamples=Samples[0]=MyTrig->ThetaSamples;
  PhiSamples=Samples[1]=MyTrig->PhiSamples;
  USamples=Samples[2]=MyTrig->USamples;
  VSamples=Samples[3]=MyTrig->VSamples;

  My4D->Min[0]=MyTrig->Thetamin;
  My4D->Min[1]=MyTrig->Phimin;
  My4D->Min[2]=MyTrig->Umin;
  My4D->Min[3]=MyTrig->Vmin;

  My4D->Delta[0]=MyTrig->DeltaTheta;
  My4D->Delta[1]=MyTrig->DeltaPhi;
  My4D->Delta[2]=MyTrig->DeltaU;
  My4D->Delta[3]=MyTrig->DeltaV;

  outsig=My4D->Signal;
  if (InVolume->Dim!=3)
    Error("Input MultiStruct has wrong dimension (NumericalRadon3D)");

  sig=InVolume->Signal;
  Xmin=InVolume->Min[0];
  Ymin=InVolume->Min[1];
  Zmin=InVolume->Min[2];

  DeltaX=InVolume->Delta[0];
  DeltaY=InVolume->Delta[1];
  DeltaZ=InVolume->Delta[2];

  XSamples=InVolume->Samples[0];
  YSamples=InVolume->Samples[1];
  ZSamples=InVolume->Samples[2];
  
  /*trigoff=0;*/

  ar_tau_x=MyTrig->tau_x;
  ar_tau_y=MyTrig->tau_y;
  ar_tau_z=MyTrig->tau_z;
  ar_alpha_x=MyTrig->alpha_x;
  ar_alpha_y=MyTrig->alpha_y;
  ar_alpha_z=MyTrig->alpha_z;
  ar_beta_x=MyTrig->beta_x;
  ar_beta_y=MyTrig->beta_y;
  ar_beta_z=MyTrig->beta_z;

#ifdef SGI  
#pragma parallel
#pragma local(trigoff,tau_x,tau_y,tau_z,alpha_x,alpha_y,alpha_z)
#pragma local(atau_x,atau_y,atau_z,uuint,vvint,u,v,rectau)
#pragma local(broek,ax,bx,ay,by,az,bz,rad,xint,yint,zint,xsint,ysint,zsint)
#pragma local(beta_x,beta_y,beta_z)
#pragma byvalue(PhiSamples,ThetaSamples,ar_tau_x,ar_tau_y,ar_tau_z)
#pragma byvalue(ar_alpha_x,ar_alpha_y,ar_alpha_z)
#pragma byvalue(ar_beta_x,ar_beta_y,ar_beta_z)
#pragma byvalue(USamples,VSamples,MyTrig,DeltaX,DeltaY,DeltaZ)
#pragma byvalue(Xmin,Ymin,Zmin,XSamples,YSamples,ZSamples)
#pragma byvalue(sig,outsig)
#endif

  {
  /*
  for (phiint=0;phiint<PhiSamples;phiint++)
  {
    Print(_DNormal,"%3i (%3i) (NumericalRadon3D)\r",
          phiint,MyTrig->PhiSamples);
    for (thetaint=0;thetaint<ThetaSamples;thetaint++)
  */

  #pragma pfor iterate(trigoff=0;(PhiSamples*ThetaSamples);1)
  for(trigoff=0; trigoff<PhiSamples*ThetaSamples;trigoff++)
    {
      Print(_DNormal,"%3i (%3i) (NumericalRadon3D)\r", trigoff,PhiSamples*ThetaSamples);
  
      tau_x=ar_tau_x[trigoff];
      tau_y=ar_tau_y[trigoff];
      tau_z=ar_tau_z[trigoff];
      
      alpha_x=ar_alpha_x[trigoff];
      alpha_y=ar_alpha_y[trigoff];
      alpha_z=ar_alpha_z[trigoff];
     
      beta_x=ar_beta_x[trigoff];
      beta_y=ar_beta_y[trigoff];
      beta_z=ar_beta_z[trigoff];

      atau_x=fabs(tau_x);  
      atau_y=fabs(tau_y);
      atau_z=fabs(tau_z);
      
      for (uuint=0;uuint<USamples;uuint++)
      {
	u=MyTrig->u[uuint];
	for (vvint=0;vvint<VSamples;vvint++)
	{
	  v=MyTrig->v[vvint];
          if ((atau_x>atau_y) && (atau_x>atau_z))
	  {
           /* Projecting onto the x-axis */
            rectau=DeltaX/atau_x;
	    broek=tau_y/tau_x;

	    ay=DeltaX/DeltaY*broek;
	    by=(broek*Xmin
		+u*(alpha_y-broek*alpha_x)
		+v*(beta_y-broek*beta_x)
		-Ymin)/DeltaY+0.5;
	    broek=tau_z/tau_x;
	    az=DeltaX/DeltaZ*broek;
	    bz=(broek*Xmin
		+u*(alpha_z-broek*alpha_x)
		+v*(beta_z-broek*beta_x)
		-Zmin)/DeltaZ+0.5;
	    rad=0.0;
	    for (xint=0;xint<XSamples;xint++)
	    {
	      ysint=ay*xint+by;
	      zsint=az*xint+bz;
              if ((ysint>=0) && (zsint>=0))
              {
                yint=(int)ysint;
                zint=(int)zsint;
                if ((yint<YSamples) && (zint<ZSamples))
                  rad+=sig[xint+XSamples*(yint+YSamples*zint)];
              }
	    }      
	  }
	  else
	    if (atau_y>atau_z)
	    {
	      /* Projecting onto the y-axis */
              rectau=DeltaY/atau_y;
	      broek=tau_x/tau_y;
	      ax=DeltaY/DeltaX*broek;
	      bx=(broek*Ymin
		  +u*(alpha_x-broek*alpha_y)
		  +v*(beta_x-broek*beta_y)
		  -Xmin)/DeltaX+0.5;
	      
	      broek=tau_z/tau_y;
	      az=DeltaY/DeltaZ*broek;
	      bz=(broek*Ymin
		  +u*(alpha_z-broek*alpha_y)
		  +v*(beta_z-broek*beta_y)
		  -Zmin)/DeltaZ+0.5;
	      rad=0.0;
	      for (yint=0;yint<YSamples;yint++)
	      {
		xsint=ax*yint+bx;
		zsint=az*yint+bz;
		if ((xsint>=0) && (zsint>0))
                {
                  xint=(int)xsint;
                  zint=(int)zsint;
                  if ((xint<XSamples) && (zint<ZSamples))
                    rad+=sig[xint+XSamples*(yint+YSamples*zint)];
                }
	      }
	    }
	    else
	    {
	      /* Projecting onto the z-axis */
              rectau=DeltaZ/atau_z;
	      broek=tau_x/tau_z;
	      ax=DeltaZ/DeltaX*broek;
	      bx=(broek*Zmin
		  +u*(alpha_x-broek*alpha_z)
		  +v*(beta_x-broek*beta_z)
		  -Xmin)/DeltaX+0.5;
	      
	      broek=tau_y/tau_z;
	      ay=DeltaZ/DeltaY*broek;
	      by=(broek*Zmin
		  +u*(alpha_y-broek*alpha_z)
		  +v*(beta_y-broek*beta_z)
		  -Ymin)/DeltaY+0.5;
	      rad=0.0;
	      for (zint=0;zint<ZSamples;zint++)
	      {
		xsint=ax*zint+bx;
		ysint=ay*zint+by;
                if ((xsint>=0) && (ysint>0))
                {
                  xint=(int)xsint;
                  yint=(int)ysint;
                  if ((xint<XSamples) && (yint<YSamples))
                    rad+=sig[xint+XSamples*(yint+YSamples*zint)];
                }
	      }
	    }
	  outsig[trigoff+ThetaSamples*(PhiSamples*
                                     (uuint+USamples*vvint))]=rad*rectau; 
	}
      }
    }
  } /* End Parallel*/
  Print(_DNormal,"Done      (NumericalRadon3D)                \n");
}


/****************************************************************
[NAME]
NumericalRadon3D

[SYNOPSIS]
 MultiStruct* NumericalRadon3D(MultiStruct *InVolume,TrigDef *MyTrig)

[DESCRIPTION]
The function will numerically calculate line integrals through
a three dimensional volume. The sampling of the Radon domain is given 
through {\tt MyTrig}.

[USAGE]
{\tt MyMultiStruct=NumericalRadon3D(InVolume,MyTrig);}

[NOTE]
See NumericalRadon3D\_Restore.

[REVISION]
24/8-95 Peter Toft (Split function into two)
****************************************************************/
MultiStruct* NumericalRadon3D(MultiStruct *InVolume,TrigDef *MyTrig)
{
  int Samples[4];
  MultiStruct *MyRadon;
  
  Samples[0]=MyTrig->ThetaSamples;
  Samples[1]=MyTrig->PhiSamples;
  Samples[2]=MyTrig->USamples;
  Samples[3]=MyTrig->VSamples;

  MyRadon=NewMultiStruct("NumericalRadon3D",4,Samples,_RealArray);

  NumericalRadon3D_Restore(MyRadon,InVolume,MyTrig);
  
  return MyRadon;
}


/****************************************************************
[NAME]
NumericalRadon3D\_OneSample\_NN

[SYNOPSIS]
 int NumericalRadon3D_OneSample_NN(VolWeight *,TrigDef *,XYZdef *,
                                  int,int,int,int)

[DESCRIPTION]

The function will numerically calculate line integrals through a three
dimensional volume ONLY at one samples in the parameter domain. The
function will number of samples and in the first parameter the indices
and weight factors corresponding to the voxels. The sampling of the
Radon domain is given through {\tt MyTrig}. {\tt uuint, vvint, thetaint}
and {\tt phiint} determines the sample in the parameter domain. The
function will return in {\tt VolWeight}, which must be externally be
allocated with at least a length of {\tt max\{XSamples,YSamples,ZSamples\}}.

[USAGE]
{\tt NoOfSamples=NumericalRadon3D\_OneSample\_NN(MyWeight,
                                               MyTrig,
                                               MyXYZ,
                                               int thetaint,
                                               int phiint
                                               int uuint,
                                               int vvint);}

[NOTE] 
Uses currently nearest neighbor interpolation. See also
NumericalRadon3D.

[REVISION]
27/6-95 Peter Toft\\
29/6-95 Peter Toft (Bug fix - rectau into DeltaX*rectau and so on)\\
10/12-95 Peter Toft - Dropped OrthoVecors and made 
use of individual vectors. See TrigDef.
****************************************************************/
int NumericalRadon3D_OneSample_NN(VolWeight *MyWeight,
                                  TrigDef *MyTrig,
                                  XYZDef *MyXYZ,
                                  int thetaint,
                                  int phiint,
                                  int uuint,
                                  int vvint)
{
  VolWeight *LocalWeight;
  float broek,u,v;
  float tau_x,tau_y,tau_z;
  float atau_x,atau_y,atau_z;
  float ax,bx,ay,by,az,bz;
  float alpha_x,alpha_y,alpha_z;
  float xsint,ysint,zsint;
  float beta_x,beta_y,beta_z;
  float DeltaX,DeltaY,DeltaZ;
  int XSamples,YSamples,ZSamples,ThetaSamples;
  int xint,yint,zint,counter;
  float Xmin,Ymin,Zmin;
  float rectau;
  int trigoff;

  ThetaSamples=MyTrig->ThetaSamples;
  counter=0;
  Xmin=MyXYZ->Xmin;
  Ymin=MyXYZ->Ymin;
  Zmin=MyXYZ->Zmin;

  DeltaX=MyXYZ->DeltaX;
  DeltaY=MyXYZ->DeltaY;
  DeltaZ=MyXYZ->DeltaZ;

  XSamples=MyXYZ->XSamples;
  YSamples=MyXYZ->YSamples;
  ZSamples=MyXYZ->ZSamples;
  
  trigoff=thetaint+phiint*ThetaSamples;

  tau_x=MyTrig->tau_x[trigoff];
  tau_y=MyTrig->tau_y[trigoff];
  tau_z=MyTrig->tau_z[trigoff];
  alpha_x=MyTrig->alpha_x[trigoff];
  alpha_y=MyTrig->alpha_y[trigoff];
  alpha_z=MyTrig->alpha_z[trigoff];
  beta_x=MyTrig->beta_x[trigoff];
  beta_y=MyTrig->beta_y[trigoff];
  beta_z=MyTrig->beta_z[trigoff];

  atau_x=fabs(tau_x);  
  atau_y=fabs(tau_y);
  atau_z=fabs(tau_z);
        
  u=MyTrig->u[uuint];
  v=MyTrig->v[vvint];
  if ((atau_x>atau_y) && (atau_x>atau_z))
  {
    /* Projecting onto the x-axis */
    rectau=DeltaX/atau_x;
    broek=tau_y/tau_x;
    ay=DeltaX/DeltaY*broek;
    by=(broek*Xmin
        +u*(alpha_y-broek*alpha_x)
        +v*(beta_y-broek*beta_x)
        -Ymin)/DeltaY+0.5;
    broek=tau_z/tau_x;
    az=DeltaX/DeltaZ*broek;
    bz=(broek*Xmin
        +u*(alpha_z-broek*alpha_x)
        +v*(beta_z-broek*beta_x)
        -Zmin)/DeltaZ+0.5;
#pragma parallel
#pragma local(xint,yint,zint,ysint,zsint,LocalWeight)
#pragma shared(counter)
#pragma byvalue(XSamples,YSamples,ZSamples,ay,by,az,bz,MyWeight,rectau)
    {
#pragma pfor iterate(xint=0;XSamples;1)
    for (xint=0;xint<XSamples;xint++)
    {
      ysint=ay*xint+by;
      zsint=az*xint+bz;
      if ((ysint>=0) && (zsint>=0))
      {
        yint=(int)ysint;
        zint=(int)zsint;
        if ((yint<YSamples) && (zint<ZSamples)) 
        {
          LocalWeight=&MyWeight[counter];
#pragma critical
          {
          counter++;
        }
          LocalWeight->A=rectau;
          LocalWeight->VolIndex=xint+XSamples*(yint+YSamples*zint);
        }
      }      
    }
  }/*end Parallel*/
  }
  else
    if (atau_y>atau_z)
    {
      /* Projecting onto the y-axis */
      rectau=DeltaY/atau_y;
      broek=tau_x/tau_y;
      ax=DeltaY/DeltaX*broek;
      bx=(broek*Ymin
          +u*(alpha_x-broek*alpha_y)
          +v*(beta_x-broek*beta_y)
          -Xmin)/DeltaX+0.5;
      
      broek=tau_z/tau_y;
      az=DeltaY/DeltaZ*broek;
      bz=(broek*Ymin
          +u*(alpha_z-broek*alpha_y)
          +v*(beta_z-broek*beta_y)
          -Zmin)/DeltaZ+0.5;

#pragma parallel
#pragma local(xint,yint,zint,zsint,xsint,LocalWeight)
#pragma shared(counter)
#pragma byvalue(XSamples,YSamples,ZSamples,ax,bx,az,bz,MyWeight,rectau)
    {
#pragma pfor iterate(yint=0;YSamples;1)
      for (yint=0;yint<YSamples;yint++)
      {
        xsint=ax*yint+bx;
        zsint=az*yint+bz;
        if ((xsint>=0) && (zsint>0))
        {
          xint=(int)xsint;
          zint=(int)zsint;
          if ((xint<XSamples) && (zint<ZSamples))
          {
            LocalWeight=&MyWeight[counter];
#pragma critical
            {
              counter++;
            }       
            LocalWeight->A=rectau;
            LocalWeight->VolIndex=xint+XSamples*(yint+YSamples*zint);
          }
        }
      }
      }/* End Parallel */
    }
    else
    {
      /* Projecting onto the z-axis */
      rectau=DeltaZ/atau_z;
      broek=tau_x/tau_z;
      ax=DeltaZ/DeltaX*broek;
      bx=(broek*Zmin
          +u*(alpha_x-broek*alpha_z)
          +v*(beta_x-broek*beta_z)
          -Xmin)/DeltaX+0.5;
	      
      broek=tau_y/tau_z;
      ay=DeltaZ/DeltaY*broek;
      by=(broek*Zmin
          +u*(alpha_y-broek*alpha_z)
          +v*(beta_y-broek*beta_z)
          -Ymin)/DeltaY+0.5;
#pragma parallel
#pragma local(xint,yint,zint,ysint,xsint,LocalWeight)
#pragma shared(counter)
#pragma byvalue(XSamples,YSamples,ZSamples,ay,by,ax,bx,MyWeight,rectau)
    {
#pragma pfor iterate(zint=0;ZSamples;1)
    for (zint=0;zint<ZSamples;zint++)
      {
        xsint=ax*zint+bx;
        ysint=ay*zint+by;
        if ((xsint>=0) && (ysint>0))
        {
          xint=(int)xsint;
          yint=(int)ysint;
          if ((xint<XSamples) && (yint<YSamples))
          {
            LocalWeight=&MyWeight[counter];
#pragma critical
            {
              counter++;
            }
            LocalWeight->A=rectau;
            LocalWeight->VolIndex=xint+XSamples*(yint+YSamples*zint);
          }
        }
      }
    } /* End Parallel */
    }

  return counter;
}


/****************************************************************
[NAME]
NumericalRadon3D\_OneSample\_BL

[SYNOPSIS]
 int NumericalRadon3D_OneSample_BL(VolWeight *,TrigDef *,XYZdef *,
                                  int,int,int,int)
  
[DESCRIPTION]

The function will numerically calculate line integrals through a three
dimensional volume ONLY at one samples in the parameter domain. The
function will number of samples and in the first parameter the indices
and weight factors corresponding to the voxels. The sampling of the
Radon domain is given through {\tt MyTrig}. {\tt uuint, vvint, thetaint}
and {\tt phiint} determines the sample in the parameter domain. The
function will return in {\tt VolWeight}, which must be externally be
allocated with at least a length of {\tt max\{XSamples,YSamples,ZSamples\}}.

[USAGE]
{\tt NoOfSamples=NumericalRadon3D\_OneSample\_BL(MyWeight,
                                               MyTrig,
                                               MyXYZ,
                                               int thetaint,
                                               int phiint
                                               int uuint,
                                               int vvint);}

[NOTE] 
Uses bilinear interpolation. See also NumericalRadon3D.

[REVISION]
29/6-95 Peter Toft\\
29/6-95 Peter Toft (Bug fix - rectau into DeltaX*rectau and so on)\\
10/12-95 Peter Toft - Dropped OrthoVecors and made 
use of individual vectors. See TrigDef.
****************************************************************/
int NumericalRadon3D_OneSample_BL(VolWeight *MyWeight,
                                  TrigDef *MyTrig,
                                  XYZDef *MyXYZ,
                                  int thetaint,
                                  int phiint,
                                  int uuint,
                                  int vvint)
{
  VolWeight *LocalWeight;
  float broek,u,v;
  float tau_x,tau_y,tau_z;
  float atau_x,atau_y,atau_z;
  float ax,bx,ay,by,az,bz;
  float dx,dy,dz;
  float alpha_x,alpha_y,alpha_z;
  float xsint,ysint,zsint;
  float beta_x,beta_y,beta_z;
  float DeltaX,DeltaY,DeltaZ;
  int XSamples,YSamples,ZSamples;
  int XSamplesM1,YSamplesM1,ZSamplesM1,VolIndex;
  int XYSamples;
  int trigoff,ThetaSamples;
  int xint,yint,zint,counter;
  float Xmin,Ymin,Zmin;
  float rectau;

  counter=0;
  Xmin=MyXYZ->Xmin;
  Ymin=MyXYZ->Ymin;
  Zmin=MyXYZ->Zmin;

  DeltaX=MyXYZ->DeltaX;
  DeltaY=MyXYZ->DeltaY;
  DeltaZ=MyXYZ->DeltaZ;

  XSamples=MyXYZ->XSamples;
  YSamples=MyXYZ->YSamples;
  ZSamples=MyXYZ->ZSamples;

  XSamplesM1=XSamples-1;
  YSamplesM1=YSamples-1;
  ZSamplesM1=ZSamples-1;
  
  XYSamples=XSamples*YSamples;
  ThetaSamples=MyTrig->ThetaSamples;

  trigoff=thetaint+phiint*ThetaSamples;

  tau_x=MyTrig->tau_x[trigoff];
  tau_y=MyTrig->tau_y[trigoff];
  tau_z=MyTrig->tau_z[trigoff];
  alpha_x=MyTrig->alpha_x[trigoff];
  alpha_y=MyTrig->alpha_y[trigoff];
  alpha_z=MyTrig->alpha_z[trigoff];
  beta_x=MyTrig->beta_x[trigoff];
  beta_y=MyTrig->beta_y[trigoff];
  beta_z=MyTrig->beta_z[trigoff];

  atau_x=fabs(tau_x);  
  atau_y=fabs(tau_y);
  atau_z=fabs(tau_z);
        
  u=MyTrig->u[uuint];
  v=MyTrig->v[vvint];

  if ((atau_x>atau_y) && (atau_x>atau_z))
  {
    /* Projecting onto the x-axis */
    rectau=DeltaX/atau_x;
    broek=tau_y/tau_x;
    ay=DeltaX/DeltaY*broek;
    by=(broek*Xmin
        +u*(alpha_y-broek*alpha_x)
        +v*(beta_y-broek*beta_x)
        -Ymin)/DeltaY;
    broek=tau_z/tau_x;
    az=DeltaX/DeltaZ*broek;
    bz=(broek*Xmin
        +u*(alpha_z-broek*alpha_x)
        +v*(beta_z-broek*beta_x)
        -Zmin)/DeltaZ;
    for (xint=0;xint<XSamples;xint++)
    {
      ysint=ay*xint+by;
      zsint=az*xint+bz;
      if ((ysint>=0) && (zsint>=0))
      {
        yint=(int)ysint;
        zint=(int)zsint;
        if ((yint<YSamplesM1) && (zint<ZSamplesM1)) 
        {
          dy=ysint-yint;
          dz=zsint-zint;
          VolIndex=xint+XSamples*(yint+YSamples*zint);

          /* Y=Lower Z=Lower */
          LocalWeight=&MyWeight[counter++];
          LocalWeight->A=rectau*(1-dy)*(1-dz);
          LocalWeight->VolIndex=VolIndex;

          /* Y=Higher Z=Lower */
          LocalWeight=&MyWeight[counter++];
          LocalWeight->A=rectau*dy*(1-dz);
          LocalWeight->VolIndex=VolIndex+XSamples;

          /* Y=Lower Z=Higher */
          LocalWeight=&MyWeight[counter++];
          LocalWeight->A=rectau*(1-dy)*dz;
          LocalWeight->VolIndex=VolIndex+XYSamples;
 
          /* Y=Higher Z=Higher */
          LocalWeight=&MyWeight[counter++];
          LocalWeight->A=rectau*dy*dz;
          LocalWeight->VolIndex=VolIndex+XSamples+XYSamples;
        }
      }      
    }
  }
  else
    if (atau_y>atau_z)
    {
      /* Projecting onto the y-axis */
      rectau=DeltaY/atau_y;
      broek=tau_x/tau_y;
      ax=DeltaY/DeltaX*broek;
      bx=(broek*Ymin
          +u*(alpha_x-broek*alpha_y)
          +v*(beta_x-broek*beta_y)
          -Xmin)/DeltaX;
      
      broek=tau_z/tau_y;
      az=DeltaY/DeltaZ*broek;
      bz=(broek*Ymin
          +u*(alpha_z-broek*alpha_y)
          +v*(beta_z-broek*beta_y)
          -Zmin)/DeltaZ;
      for (yint=0;yint<YSamples;yint++)
      {
        xsint=ax*yint+bx;
        zsint=az*yint+bz;
        if ((xsint>=0) && (zsint>0))
        {
          xint=(int)xsint;
          zint=(int)zsint;
          if ((xint<XSamplesM1) && (zint<ZSamplesM1))
          {
            dx=xsint-xint;
            dz=zsint-zint;
            VolIndex=xint+XSamples*(yint+YSamples*zint);

            /* X=Lower Z=Lower */
            LocalWeight=&MyWeight[counter++];
            LocalWeight->A=rectau*(1-dx)*(1-dz);
            LocalWeight->VolIndex=VolIndex;

            /* X=Higher Z=Lower */
            LocalWeight=&MyWeight[counter++];
            LocalWeight->A=rectau*dx*(1-dz);
            LocalWeight->VolIndex=VolIndex+1;

            /* X=Lower Z=Higher */
            LocalWeight=&MyWeight[counter++];
            LocalWeight->A=rectau*(1-dx)*dz;
            LocalWeight->VolIndex=VolIndex+XYSamples;
 
            /* X=Higher Z=Higher */
            LocalWeight=&MyWeight[counter++];
            LocalWeight->A=rectau*dx*dz;
            LocalWeight->VolIndex=VolIndex+1+XYSamples;
          }
        }
      }
    }
    else
    {
      /* Projecting onto the z-axis */
      rectau=DeltaZ/atau_z;
      broek=tau_x/tau_z;
      ax=DeltaZ/DeltaX*broek;
      bx=(broek*Zmin
          +u*(alpha_x-broek*alpha_z)
          +v*(beta_x-broek*beta_z)
          -Xmin)/DeltaX;
	      
      broek=tau_y/tau_z;
      ay=DeltaZ/DeltaY*broek;
      by=(broek*Zmin
          +u*(alpha_y-broek*alpha_z)
          +v*(beta_y-broek*beta_z)
          -Ymin)/DeltaY;
      for (zint=0;zint<ZSamples;zint++)
      {
        xsint=ax*zint+bx;
        ysint=ay*zint+by;
        if ((xsint>=0) && (ysint>0))
        {
          xint=(int)xsint;
          yint=(int)ysint;
          if ((xint<XSamplesM1) && (yint<YSamplesM1))
          {
            dx=xsint-xint;
            dy=ysint-yint;
            VolIndex=xint+XSamples*(yint+YSamples*zint);

            /* X=Lower Y=Lower */
            LocalWeight=&MyWeight[counter++];
            LocalWeight->A=rectau*(1-dx)*(1-dy);
            LocalWeight->VolIndex=VolIndex;

            /* X=Higher Y=Lower */
            LocalWeight=&MyWeight[counter++];
            LocalWeight->A=rectau*dx*(1-dy);
            LocalWeight->VolIndex=VolIndex+1;

            /* X=Lower Y=Higher */
            LocalWeight=&MyWeight[counter++];
            LocalWeight->A=rectau*(1-dx)*dy;
            LocalWeight->VolIndex=VolIndex+XSamples;
 
            /* X=Higher Y=Higher */
            LocalWeight=&MyWeight[counter++];
            LocalWeight->A=rectau*dx*dy;
            LocalWeight->VolIndex=VolIndex+1+XSamples;
          }
        }
      }
    }

  return counter;
}


/****************************************************************
[NAME]
BackProject3D\_Restore

[SYNOPSIS]
 void BackProject3D_Restore(MultiStruct *MyVol,
                            MultiStruct *My4D,
                            XYZDef *MyXYZ,
                            TrigDef *MyTrig)

[DESCRIPTION] 
The function will backproject a 4D set of line integrals through
a 3D volume.

\begin{equation}
\tilde{g}(x,y,z)=
\int_{\theta}\int_{\phi} 
\check{g} (\theta,\phi,u=\mb{r}\cdot\mb{\alpha},v=\mb{r}\cdot\mb{\beta})
~d\phi~\theta
\end{equation}

[USAGE]
{\tt BackProject3D\_Restore(MyVolume,My4D,MyXYZ,MyTrig);}

[NOTE]
This function requires that the volume has been allocated elsewhere.

[REVISION]
20/5-95 Peter Toft\\
31/5-95 Peter and Peter.(Bilinear interpol, cos(phi) factor inc.)\\
14/6-95 Peter Toft. Faster (floor)\\
19/6-95 Peter Toft. Optimization. (more local vars)\\
24/8-95 Peter Toft. Split into two functions.\\
10/12-95 Peter Toft - Dropped OrthoVecors and made 
use of individual vectors. See TrigDef.
****************************************************************/
void BackProject3D_Restore(MultiStruct *MyVol,
                           MultiStruct *My4D,
                           XYZDef *MyXYZ,
                           TrigDef *MyTrig)
{
  int phiint,xint,yint,zint,uuint,vvint;
  int thetaphiint;
  int offXYZ,SigOff,MultThetaPhi,MultThetaPhiU,MultThetaPhiUp1;
  int XSamples,YSamples,ZSamples,USamples,VSamples,PhiSamples,ThetaSamples;
  int USamplesm1,VSamplesm1;
  float sum,trigfactor;
  float x,y,z,deltau,deltav;
  float usint,vsint,recDeltaU,recDeltaV,Umin,Vmin;
  float *sig;
  float *cosphi;
  float *ar_alpha_x,*ar_alpha_y,*ar_alpha_z;
  float *ar_beta_x,*ar_beta_y,*ar_beta_z;
  float *ar_tau_x,*ar_tau_y,*ar_tau_z;

  Print(_DDebug,"3D Backprojecting...   (BackProject3D_Restore)\n");

  if (MyVol->Dim!=3)
    Error("Volume has wrong number of dimensions (BackProject3D_Restore)\n");

  if (My4D->Dim!=4)
    Error("Sinogram has wrong number of dimensions (BackProject3D_Restore)\n");

  XSamples=MyXYZ->XSamples;
  YSamples=MyXYZ->YSamples;
  ZSamples=MyXYZ->ZSamples;

  USamples=MyTrig->USamples;
  VSamples=MyTrig->VSamples;
  PhiSamples=MyTrig->PhiSamples;
  ThetaSamples=MyTrig->ThetaSamples;
  USamplesm1=USamples-1;
  VSamplesm1=VSamples-1;
 
  /*
  cosphi=FloatVector(PhiSamples);
  for (phiint=0;phiint<PhiSamples;phiint++)
    cosphi[phiint]=cos(MyTrig->phi[phiint]);
    */
  trigfactor=MyTrig->DeltaTheta*MyTrig->DeltaPhi;

  ar_tau_x=MyTrig->tau_x;
  ar_tau_y=MyTrig->tau_y;
  ar_tau_z=MyTrig->tau_z;
  ar_alpha_x=MyTrig->alpha_x;
  ar_alpha_y=MyTrig->alpha_y;
  ar_alpha_z=MyTrig->alpha_z;
  ar_beta_x=MyTrig->beta_x;
  ar_beta_y=MyTrig->beta_y;
  ar_beta_z=MyTrig->beta_z;


  Umin=MyTrig->Umin;
  Vmin=MyTrig->Vmin;
  recDeltaU=1.0/MyTrig->DeltaU;
  recDeltaV=1.0/MyTrig->DeltaV;

  MyVol->Min[0]=MyXYZ->Xmin;
  MyVol->Min[1]=MyXYZ->Ymin;
  MyVol->Min[2]=MyXYZ->Zmin;
  MyVol->Delta[0]=MyXYZ->DeltaX;
  MyVol->Delta[1]=MyXYZ->DeltaY;
  MyVol->Delta[2]=MyXYZ->DeltaZ;

  MultThetaPhi=PhiSamples*ThetaSamples;
  MultThetaPhiU=PhiSamples*ThetaSamples*USamples;
  MultThetaPhiUp1=PhiSamples*ThetaSamples*(1+USamples);

  sig=My4D->Signal;

#pragma parallel
#pragma local(zint,z,yint,y,xint,x,offXYZ,sum,thetaphiint)
#pragma local(usint,vsint,uuint,vvint,SigOff,deltau,deltav,cosphi,phiint)
#pragma shared(MyTrig)
#pragma byvalue(ZSamples,YSamples,XSamples,MultThetaPhi,PhiSamples)
#pragma byvalue(recDeltaU,recDeltaV,MyXYZ,Umin,Vmin)
#pragma byvalue(USamplesm1,VSamplesm1,ar_alpha_x,ar_alpha_y,ar_alpha_z)
#pragma byvalue(ar_beta_x,ar_beta_y,ar_beta_z)
#pragma byvalue(sig,MultThetaPhi,MultThetaPhiU,MultThetaPhiUp1)
#pragma byvalue(USamples,trigfactor,MyVol,ThetaSamples)
  {
  cosphi=FloatVector(PhiSamples);
  for (phiint=0;phiint<PhiSamples;phiint++)
    cosphi[phiint]=cos(MyTrig->phi[phiint]);
    
#pragma pfor iterate (zint=0;ZSamples;1)
    for (zint=0;zint<ZSamples;zint++)
      {
	Print(_DNormal,
		  "3D Backprojecting: %3i(%3i) iterations done\r",zint,ZSamples);
        z=MyXYZ->z[zint];
        for (yint=0;yint<YSamples;yint++)
          {
            y=MyXYZ->y[yint];
            offXYZ=XSamples*(yint+zint*YSamples);
            for (xint=0;xint<XSamples;xint++)
              {
                x=MyXYZ->x[xint];
                sum=0.0;
                for (thetaphiint=0;thetaphiint<MultThetaPhi;thetaphiint++)
                  {
                    usint=recDeltaU*(ar_alpha_x[thetaphiint]*x+
                                     ar_alpha_y[thetaphiint]*y+
                                     ar_alpha_z[thetaphiint]*z-Umin);
                    vsint=recDeltaV*(ar_beta_x[thetaphiint]*x+
                                     ar_beta_y[thetaphiint]*y+
                                     ar_beta_z[thetaphiint]*z-Vmin);
                    
                    if ((usint>=0) && (vsint>=0))
                      {
                        uuint=(int)usint;
                        vvint=(int)vsint;
                        if ((uuint<USamplesm1) && (vvint<VSamplesm1))
                          {
                            SigOff=thetaphiint+MultThetaPhi*(uuint+USamples*vvint);
                            
                            deltau=usint-uuint;
                            deltav=vsint-vvint;
                            
                            sum+=((1-deltau)*((1-deltav)*sig[SigOff]
                                              +deltav*sig[SigOff+MultThetaPhiU])
                                  +deltau*((1-deltav)*sig[SigOff+MultThetaPhi]
                                           +deltav*sig[SigOff+MultThetaPhiUp1]))*
                              cosphi[(int)(thetaphiint/ThetaSamples)];
                          }
                      }
                  }
                #pragma critical
                MyVol->Signal[xint+offXYZ]=sum*trigfactor; 
              }
          }
      }
  Free(cosphi);
  } /*End parallel*/
  /*Free(cosphi);*/
  Print(_DNormal,"Done      (BackProject3D_Restore)                  \n");
}


/****************************************************************
[NAME]
BackProject3D

[SYNOPSIS]
 MultiStruct *BackProject3D(MultiStruct *My4D,XYZDef *MyXYZ,TrigDef *MyTrig)

[DESCRIPTION] 
The function will backproject a 4D set of line integrals through
a 3D volume.

\begin{equation}
\tilde{g}(x,y,z)=
\int_{\theta}\int_{\phi} 
\check{g} (\theta,\phi,u=\mb{r}\cdot\mb{\alpha},v=\mb{r}\cdot\mb{\beta})
~d\phi~\theta
\end{equation}

[USAGE]
{\tt My3D=BackProject3D(My4D,MyXYZ,MyTrig);}

[NOTE]
See also {\tt BackProject3D\_Restore}

[REVISION]
20/5-95 Peter Toft\\
31/5-95 Peter and Peter.(Bilinear interpol, cos(phi) factor inc.)\\
14/6-95 Peter Toft. Faster (floor)\\
19/6-95 Peter Toft. Optimization. (more local vars)\\
24/8-95 Peter Toft. Split in two functions.\\
****************************************************************/
MultiStruct *BackProject3D(MultiStruct *My4D,XYZDef *MyXYZ,TrigDef *MyTrig)
{
  MultiStruct *MyVol;
  int XYZSamples[3];
 
  Print(_DDebug,"3D Backprojecting... (BackProject3D) \n");

  XYZSamples[0]=MyXYZ->XSamples;
  XYZSamples[1]=MyXYZ->YSamples;
  XYZSamples[2]=MyXYZ->ZSamples;

  MyVol=NewMultiStruct("Backproject",3,XYZSamples,_RealArray);

  BackProject3D_Restore(MyVol,My4D,MyXYZ,MyTrig);

  return MyVol;
}








