#include <math.h>
#include <time.h>
#include <stdio.h>
#include <ctype.h>
#include <string.h>
#include <stdlib.h>
#include <stdarg.h>
#include <malloc.h>
#include "3dtools.h"
#include "Multi_IO.h"
#include "3dmisc.h"


/****************************************************************
[NAME]
WriteMultiStruct

[SYNOPSIS]
void WriteMultiStruct(MultiStruct *MyMultiStruct)

[DESCRIPTION]

The function will save the {\tt MultiStruct} to the disc under the
filename contained in the structure. A file extension {\tt .f<N>f} is
added, where {\tt <N>} is the dimension.
 
[USAGE] 
{\tt WriteMultiStruct(MyMultiStruct);}

[REVISION]
20/5-95 Peter Toft
****************************************************************/
void WriteMultiStruct(MultiStruct *MyMultiStruct)
{
  int Dim;
  FILE *OutFile;
  char OutFileName[160];

  sprintf(OutFileName,"%s.f%if",MyMultiStruct->FileName,MyMultiStruct->Dim);

  Print(_DNormal,"WriteMultiStruct: (%i dimensions)\n Writing `%s'\n",
	Dim=MyMultiStruct->Dim,OutFileName);

  if(!(OutFile=fopen(OutFileName,"wb"))) 
    Error("Error opening file: `%s'",OutFileName);

  fwrite(MyMultiStruct,sizeof(MultiStruct),1,OutFile);

  if((fwrite(MyMultiStruct->Samples,sizeof(int),Dim,OutFile))!=Dim)
    Error("Error writing to file: `%s'",OutFileName);

  if((fwrite(MyMultiStruct->Min,sizeof(float),Dim,OutFile))!=Dim)
    Error("Error writing to file: `%s'",OutFileName);

  if((fwrite(MyMultiStruct->Delta,sizeof(float),Dim,OutFile))!=Dim)
    Error("Error writing to file: `%s'",OutFileName);

  if((fwrite(MyMultiStruct->Signal,
	     sizeof(float),
	     MyMultiStruct->NoOfSamples*MyMultiStruct->ArrayType,
	     OutFile))!=MyMultiStruct->NoOfSamples*MyMultiStruct->ArrayType)
      Error("Error writing to file: `%s'",OutFileName);

  fclose(OutFile);
}


/****************************************************************
[NAME]
WriteMultiStruct\_NoSignal

[SYNOPSIS]
void WriteMultiStruct_NoSignal(MultiStruct *MyMultiStruct)

[DESCRIPTION]

The function will save the {\tt MultiStruct} to the disc under the
filename contained in the structure. A file extension {\tt .f<N>f} is
added, where {\tt <N>} is the dimension. No signal is written.
 
[USAGE] 
{\tt WriteMultiStruct\_NoSignal(MyMultiStruct);}

[REVISION]
20/5-95 Peter Toft
****************************************************************/
void WriteMultiStruct_NoSignal(MultiStruct *MyMultiStruct)
{
  int Dim;
  FILE *OutFile;
  char OutFileName[160],temp[8];

  strcpy(OutFileName,MyMultiStruct->FileName);
  sprintf(temp,".f%if",MyMultiStruct->Dim);
  strcat(OutFileName,temp);

  Print(_DNormal,"WriteMultiStruct_NoSignal: Writing `%s' (%i dimensions) \n",
	OutFileName,Dim=MyMultiStruct->Dim);

  if(!(OutFile=fopen(OutFileName,"wb"))) 
    Error("Error opening file: `%s'",OutFileName);

  fwrite(MyMultiStruct,sizeof(MultiStruct),1,OutFile);

  if((fwrite(MyMultiStruct->Samples,sizeof(int),Dim,OutFile))!=Dim)
    Error("Error writing to file: `%s'",OutFileName);

  if((fwrite(MyMultiStruct->Min,sizeof(float),Dim,OutFile))!=Dim)
    Error("Error writing to file: `%s'",OutFileName);

  if((fwrite(MyMultiStruct->Delta,sizeof(float),Dim,OutFile))!=Dim)
    Error("Error writing to file: `%s'",OutFileName);
  fclose(OutFile);
}


/****************************************************************
[NAME]
WriteMultiStruct\_ASCII

[SYNOPSIS]
void WriteMultiStruct_ASCII(MultiStruct *MyMultiStruct)

[DESCRIPTION]

The function will save the {\tt MultiStruct} to the disc under the
filename contained in the structure. A file extension {\tt .a<N>f} is
added, where {\tt <N>} is the dimension.
 
[USAGE] 
{\tt WriteMultiStruct\_ASCII(MyMultiStruct);}

[REVISION]
21/6-95 Peter Philipsen
****************************************************************/
void WriteMultiStruct_ASCII(MultiStruct *MyMultiStruct)
{
  int m,Dim;
  FILE *OutFile;
  char OutFileName[160];

  sprintf(OutFileName,"%s.a%if",MyMultiStruct->FileName,MyMultiStruct->Dim);

  Print(_DNormal,"WriteMultiStruct: Writing `%s' (%i dimensions) \n",
	OutFileName,Dim=MyMultiStruct->Dim);

  if(!(OutFile=fopen(OutFileName,"w"))) 
    Error("Error opening file: `%s'",OutFileName);

  /* Write Header*/
  if((fprintf(OutFile,"#MultiStruct\n"))<0)
    Error("Error writing to file: `%s'",OutFileName);
  
  if((fprintf(OutFile,"%i\n",MyMultiStruct->Id))<0) 
    Error("Error writing to file: `%s'",OutFileName);
 
  if((fprintf(OutFile,"%i\n",MyMultiStruct->LengthOfHeader))<0) 
    Error("Error writing to file: `%s'",OutFileName);
  if((fprintf(OutFile,"%i\n",MyMultiStruct->Dim))<0) 
    Error("Error writing to file: `%s'",OutFileName);
  if((fprintf(OutFile,"%i\n",MyMultiStruct->NoOfSamples))<0) 
    Error("Error writing to file: `%s'",OutFileName);
  if((fprintf(OutFile,"%i\n",MyMultiStruct->Type))<0) 
    Error("Error writing to file: `%s'",OutFileName);
  if((fprintf(OutFile,"%s\n",MyMultiStruct->FileName))<0) 
    Error("Error writing to file: `%s'",OutFileName);
  if((fprintf(OutFile,"%s\n",MyMultiStruct->Description))<0) 
    Error("Error writing to file: `%s'",OutFileName);
  if((fprintf(OutFile,"%s\n",MyMultiStruct->Date))<0) 
    Error("Error writing to file: `%s'",OutFileName);
  if((fprintf(OutFile,"%i\n",MyMultiStruct->ArrayType))<0) 
    Error("Error writing to file: `%s'",OutFileName);

  if((fprintf(OutFile,"#Dim\n"))<0) 
    Error("Error writing to file: `%s'",OutFileName);
  for (m=0;m<Dim;m++) 
    fprintf(OutFile,"%i\n",(MyMultiStruct->Samples[m]));
    
  if((fprintf(OutFile,"#Min\n"))<0) 
    Error("Error writing to file: `%s'",OutFileName);
  for (m=0;m<Dim;m++) 
    fprintf(OutFile,"%e\n",(MyMultiStruct->Min[m]));
    
  if((fprintf(OutFile,"#Delta\n"))<0) 
    Error("Error writing to file: `%s'",OutFileName);
  for (m=0;m<Dim;m++) 
    fprintf(OutFile,"%e\n",(MyMultiStruct->Delta[m]));

  if((fprintf(OutFile,"#Signal\n"))<0) 
    Error("Error writing to file: `%s'",OutFileName);
  for (m=0;m<MyMultiStruct->NoOfSamples*MyMultiStruct->ArrayType;m++) 
    fprintf(OutFile,"%e\n",(MyMultiStruct->Signal[m]));

  if((fprintf(OutFile,"#end\n"))<0) 
    Error("Error writing to file: `%s'",OutFileName);
 
  fclose(OutFile);
}


/****************************************************************
[NAME]
WriteMultiStruct\_ASCII\_NoSignal

[SYNOPSIS]
void WriteMultiStruct_ASCII_NoSignal(MultiStruct *MyMultiStruct)

[DESCRIPTION]

The function will save the {\tt MultiStruct} to the disc under the
filename contained in the structure. A file extension {\tt .a<N>f} is
added, where {\tt <N>} is the dimension. No signal is written.
 
[USAGE] 
{\tt WriteMultiStruct\_ASCII\_NoSignal(MyMultiStruct);}

[REVISION]
20/5-95 Peter Toft
****************************************************************/
void WriteMultiStruct_ASCII_NoSignal(MultiStruct *MyMultiStruct)
{
  int m,Dim;
  FILE *OutFile;
  char OutFileName[160],temp[8];

  strcpy(OutFileName,MyMultiStruct->FileName);
  sprintf(temp,".a%if",MyMultiStruct->Dim);
  strcat(OutFileName,temp);

  Print(_DNormal,"WriteMultiStruct_NoSignal: Writing `%s' (%i dimensions) \n",
	OutFileName,Dim=MyMultiStruct->Dim);

  if(!(OutFile=fopen(OutFileName,"w"))) 
    Error("Error opening file: `%s'",OutFileName);
  /* Write Header*/
  if((fprintf(OutFile,"#MultiStruct\n"))<0)
    Error("Error writing to file: `%s'",OutFileName);
  
  if((fprintf(OutFile,"%i\n",MyMultiStruct->Id))<0) 
     Error("Error writing to file: `%s'",OutFileName);
   
  if((fprintf(OutFile,"%i\n",MyMultiStruct->LengthOfHeader))<0) 
    Error("Error writing to file: `%s'",OutFileName);
  if((fprintf(OutFile,"%i\n",MyMultiStruct->Dim))<0) 
    Error("Error writing to file: `%s'",OutFileName);
  if((fprintf(OutFile,"%i\n",MyMultiStruct->NoOfSamples))<0) 
    Error("Error writing to file: `%s'",OutFileName);
  if((fprintf(OutFile,"%i\n",MyMultiStruct->Type))<0) 
    Error("Error writing to file: `%s'",OutFileName);
  if((fprintf(OutFile,"%s\n",MyMultiStruct->FileName))<0) 
    Error("Error writing to file: `%s'",OutFileName);
  if((fprintf(OutFile,"%s\n",MyMultiStruct->Description))<0) 
    Error("Error writing to file: `%s'",OutFileName);
  if((fprintf(OutFile,"%s\n",MyMultiStruct->Date))<0) 
    Error("Error writing to file: `%s'",OutFileName);
  if((fprintf(OutFile,"%i\n",MyMultiStruct->ArrayType))<0) 
    Error("Error writing to file: `%s'",OutFileName);

  if((fprintf(OutFile,"#Dim\n"))<0) 
    Error("Error writing to file: `%s'",OutFileName);
  for (m=0;m<Dim;m++) 
    fprintf(OutFile,"%i\n",(MyMultiStruct->Samples[m]));
    
  if((fprintf(OutFile,"#Min\n"))<0) 
    Error("Error writing to file: `%s'",OutFileName);
  for (m=0;m<Dim;m++) 
    fprintf(OutFile,"%e\n",(MyMultiStruct->Min[m]));
    
  if((fprintf(OutFile,"#Delta\n"))<0) 
    Error("Error writing to file: `%s'",OutFileName);
  for (m=0;m<Dim;m++) 
    fprintf(OutFile,"%e\n",(MyMultiStruct->Delta[m]));
  
  if((fprintf(OutFile,"#Signal\n"))<0) 
    Error("Error writing to file: `%s'",OutFileName);

  fclose(OutFile);
}


/****************************************************************
[NAME]
ReadMultiStruct

[SYNOPSIS]
MultiStruct * ReadMultiStruct(char *FileName,int Dim)


[DESCRIPTION]

The function will read a {\tt MultiStruct} with filename {\tt
FileName.f<Dim>f}.

[USAGE] 
{\tt MyMultiStruct=ReadMultiStruct("My3D",3);}

This will read {\tt My3D.f3f}.

[REVISION]
20/5-95 Peter Toft
****************************************************************/
MultiStruct * ReadMultiStruct(char *FileName,int Dim)
{
  int m,DoConvert;
  FILE *InFile;
  char InFileName[100],temp[8];
  MultiStruct *MyMultiStruct;
  DoConvert=FALSE;

  strcpy(InFileName,FileName);
  sprintf(temp,".f%if",Dim);
  strcat(InFileName,temp);

  if(!(MyMultiStruct=(MultiStruct *) malloc(sizeof(MultiStruct)))) 
    Error( "Memory allocation problems (ReadMultiStruct)" );

  Print(_DNormal,"ReadMultiStruct: Reading `%s' \n",FileName);
  
  if(!(InFile=fopen(InFileName,"rb")))	
    Error("Error opening file: `%s'",InFileName);
  
  if(!(MyMultiStruct=(MultiStruct *) malloc(sizeof(MultiStruct)))) 
    Error( "Memory allocation problems (ReadMultiStruct)" );

  if(1!=fread(MyMultiStruct,sizeof(MultiStruct),1,InFile))
    Error( "Cannot read file (ReadMultiStruct)" );

  if(MyMultiStruct->Id!=_00ND) {
    DoConvert=TRUE;
    convert4((char *)&MyMultiStruct->Id);
    if(MyMultiStruct->Id!=_00ND) 
      Error("Error reading MultiStruct (Converting)\n"); 
  }
  
  /* Converting MultiStruct header.*/
  if(DoConvert) {
    convert4((char *)&MyMultiStruct->LengthOfHeader);
    convert4((char *)&MyMultiStruct->Dim);
    convert4((char *)&MyMultiStruct->NoOfSamples);
    convert4((char *)&MyMultiStruct->Type);
    convert4((char *)&MyMultiStruct->ArrayType);
  }

  if (MyMultiStruct->Dim!=Dim)
    Error("MultiStruct wrong dim. (Not supported)\n");

  MyMultiStruct->Samples=IntVector(Dim);
  MyMultiStruct->Min=FloatVector(Dim);
  MyMultiStruct->Delta=FloatVector(Dim);
   
  fread(MyMultiStruct->Samples,sizeof(int),Dim,InFile);
  fread(MyMultiStruct->Min,sizeof(float),Dim,InFile);
  fread(MyMultiStruct->Delta,sizeof(float),Dim,InFile);
  
  if(DoConvert) 
    for (m=0;m<Dim;m++) {
      convert4((char *)&(MyMultiStruct->Samples[m]));
      convert4((char *)&(MyMultiStruct->Min[m]));
      convert4((char *)&(MyMultiStruct->Delta[m]));
    }
  
  
  MyMultiStruct->Signal=
    FloatVector(MyMultiStruct->NoOfSamples*MyMultiStruct->ArrayType);
  
  fread(MyMultiStruct->Signal,sizeof(float),
	MyMultiStruct->NoOfSamples*MyMultiStruct->ArrayType,InFile);
  
  if(DoConvert) 
    for (m=0;m<MyMultiStruct->NoOfSamples*MyMultiStruct->ArrayType;m++) 
      convert4((char *)&(MyMultiStruct->Signal[m]));
  
  fclose(InFile);
  return MyMultiStruct;
}


/****************************************************************
[NAME]
ReadMultiStruct\_NoSignal

[SYNOPSIS]
MultiStruct * ReadMultiStruct_NoSignal(char *FileName,int Dim)

[DESCRIPTION]

The function will read a {\tt MultiStruct} with filename {\tt
FileName.f<Dim>f}. The signal is not read, thus can be used to read
header.

[USAGE] 
{\tt MyMultiStruct=ReadMultiStruct\_NoSignal("My3D",3);}

This will read {\tt My3D.f3f} without signal.

[REVISION]
20/5-95 Peter Toft
****************************************************************/
MultiStruct * ReadMultiStruct_NoSignal(char *FileName,int Dim)
{
  int m,DoConvert;
  FILE *InFile;
  char InFileName[100],temp[8];
  MultiStruct *MyMultiStruct;
  DoConvert=FALSE;

  strcpy(InFileName,FileName);
  sprintf(temp,".f%if",Dim);
  strcat(InFileName,temp);

  if(!(MyMultiStruct=(MultiStruct *) malloc(sizeof(MultiStruct)))) 
    Error( "Memory allocation problems (ReadMultiStruct)" );

  Print(_DNormal,"ReadMultiStruct: Reading `%s' \n",FileName);
  
  if(!(InFile=fopen(InFileName,"rb")))	
    Error("Error opening file: `%s'",InFileName);
  
  if(!(MyMultiStruct=(MultiStruct *) malloc(sizeof(MultiStruct)))) 
    Error( "Memory allocation problems (ReadMultiStruct)" );

  fread(MyMultiStruct,sizeof(MultiStruct),1,InFile);

  if(MyMultiStruct->Id!=20036) {
    DoConvert=TRUE;
    convert4((char *)&MyMultiStruct->Id);
    if(MyMultiStruct->Id!=20036) 
      Error("Error reading MultiStruct (Converting)\n"); 
  }
  
  /* Converting MultiStruct header.*/
  if(DoConvert) {
    convert4((char *)&MyMultiStruct->LengthOfHeader);
    convert4((char *)&MyMultiStruct->Dim);
    convert4((char *)&MyMultiStruct->NoOfSamples);
    convert4((char *)&MyMultiStruct->Type);
    convert4((char *)&MyMultiStruct->ArrayType);
  }

  if (MyMultiStruct->Dim!=Dim)
    Error("MultiStruct wrong dim. (Not supported)\n");

  MyMultiStruct->Samples=IntVector(Dim);
  MyMultiStruct->Min=FloatVector(Dim);
  MyMultiStruct->Delta=FloatVector(Dim);
   
  fread(MyMultiStruct->Samples,sizeof(int),Dim,InFile);
  fread(MyMultiStruct->Min,sizeof(float),Dim,InFile);
  fread(MyMultiStruct->Delta,sizeof(float),Dim,InFile);
  
  if(DoConvert) 
    for (m=0;m<Dim;m++) {
      convert4((char *)&(MyMultiStruct->Samples[m]));
      convert4((char *)&(MyMultiStruct->Min[m]));
      convert4((char *)&(MyMultiStruct->Delta[m]));
    }
    
  MyMultiStruct->Signal=NULL;
  
  fclose(InFile);
  return MyMultiStruct;
}


/****************************************************************
[NAME]
ReadMultiStruct\_ASCII

[SYNOPSIS]
MultiStruct * ReadMultiStruct_ASCII(char *FileName,int Dim)


[DESCRIPTION]

The function will read a {\tt MultiStruct} with filename {\tt
FileName.a<Dim>f}.

[USAGE] 
{\tt MyMultiStruct=ReadMultiStruct\_ASCII("My3D",3);}

This will read {\tt My3D.a3f}.

[REVISION]
20/6-95 Peter Philipsen
****************************************************************/
MultiStruct * ReadMultiStruct_ASCII(char *FileName,int Dim)
{
  int m;
  FILE *InFile;
  char InFileName[100],temp[80];
  MultiStruct *MyMultiStruct;
 
  strcpy(InFileName,FileName);
  sprintf(temp,".a%if",Dim);
  strcat(InFileName,temp);

  if(!(MyMultiStruct=(MultiStruct *) malloc(sizeof(MultiStruct)))) 
    Error( "Memory allocation problems (ReadMultiStruct)" );

  Print(_DNormal,"ReadMultiStruct: Reading `%s' \n",FileName);
  
  if(!(InFile=fopen(InFileName,"rb")))	
    Error("Error opening file: `%s'",InFileName);
  
  if(!(MyMultiStruct=(MultiStruct *) malloc(sizeof(MultiStruct)))) 
    Error( "Memory allocation problems (ReadMultiStruct)" );

  /* Read MultiStructHeader from file*/
  if((fscanf(InFile,"%s",temp))==EOF) Error("Cannot read file (ASCII)\n");
  if(strcmp(temp,"#MultiStruct"))
     Error("Error reading ASCII MultiStruct \n");
  
  if((fscanf(InFile,"%i",&MyMultiStruct->Id))==EOF) Error("Cannot read file (ASCII)\n");
  if(MyMultiStruct->Id!=20036) 
    Error("Error reading ASCII MultiStruct \n"); 
  
  if((fscanf(InFile,"%i",&MyMultiStruct->LengthOfHeader))==EOF) 
    Error("Cannot read file (ASCII)\n");
  if((fscanf(InFile,"%i",&MyMultiStruct->Dim))==EOF) 
    Error("Cannot read file (ASCII)\n");
  if((fscanf(InFile,"%i",&MyMultiStruct->NoOfSamples))==EOF) 
    Error("Cannot read file (ASCII)\n");
  if((fscanf(InFile,"%i",&MyMultiStruct->Type))==EOF) 
    Error("Cannot read file (ASCII)\n");
  if((fgets(MyMultiStruct->FileName,160,InFile))==NULL) 
    Error("Cannot read file (ASCII)\n");
  if((fgets(MyMultiStruct->Description,80,InFile))==NULL) 
    Error("Cannot read file (ASCII)\n");
  if((fgets(MyMultiStruct->Date,8,InFile))==NULL) 
    Error("Cannot read file (ASCII)\n");
  if((fscanf(InFile,"%i",&MyMultiStruct->ArrayType))==EOF) 
    Error("Cannot read file (ASCII)\n");

  if (MyMultiStruct->Dim!=Dim)
    Error("MultiStruct wrong dim. (Not supported)\n");

  MyMultiStruct->Samples=IntVector(Dim);
  MyMultiStruct->Min=FloatVector(Dim);
  MyMultiStruct->Delta=FloatVector(Dim);
   
  if((fscanf(InFile,"%s",temp))==EOF) 
    Error("Cannot read file (ASCII)\n");
  if(strcmp(temp,"#Dim"))
    Error("Error reading ASCII MultiStruct \n");
  for (m=0;m<Dim;m++) 
    fscanf(InFile,"%i",&(MyMultiStruct->Samples[m]));
    
  if((fscanf(InFile,"%s",temp))==EOF) 
    Error("Cannot read file (ASCII)\n");
  if(strcmp(temp,"#Min"))
    Error("Error reading ASCII MultiStruct \n");
  for (m=0;m<Dim;m++) 
    fscanf(InFile,"%e",&(MyMultiStruct->Min[m]));
    
  if((fscanf(InFile,"%s",temp))==EOF) 
    Error("Cannot read file (ASCII)\n");
  if(strcmp(temp,"#Delta"))
    Error("Error reading ASCII MultiStruct \n");
  for (m=0;m<Dim;m++) 
    fscanf(InFile,"%e",&(MyMultiStruct->Delta[m]));
 
  MyMultiStruct->Signal=
    FloatVector(MyMultiStruct->NoOfSamples*MyMultiStruct->ArrayType);
 
  if((fscanf(InFile,"%s",temp))==EOF) 
    Error("Cannot read file (ASCII)\n");
  if(strcmp(temp,"#Signal"))
    Error("Error reading ASCII MultiStruct \n");
  for (m=0;m<MyMultiStruct->NoOfSamples*MyMultiStruct->ArrayType;m++) 
    fscanf(InFile,"%e",&(MyMultiStruct->Signal[m]));

  if((fscanf(InFile,"%s",temp))==EOF) 
    Error("Cannot read file (ASCII)\n");
  if(strcmp(temp,"#end"))
    Error("Error reading ASCII MultiStruct \n");
  
  fclose(InFile);
  return MyMultiStruct;
}


/****************************************************************
[NAME]
ReadMultiStruct\_ASCII\_NoSignal

[SYNOPSIS]
MultiStruct * ReadMultiStruct_ASCII_NoSignal(char *FileName,int Dim)

[DESCRIPTION]

The function will read a {\tt MultiStruct} with filename {\tt
FileName.a<Dim>f}. The signal is not read, thus can be used to read
header.

[USAGE] 
{\tt MyMultiStruct=ReadMultiStruct\_ASCII\_NoSignal("My3D",3);}

This will read {\tt My3D.a3f} without signal.

[REVISION]
20/6-95 Peter Philipsen
****************************************************************/
MultiStruct * ReadMultiStruct_ASCII_NoSignal(char *FileName,int Dim)
{
  int m;
  FILE *InFile;
  char InFileName[100],temp[80];
  MultiStruct *MyMultiStruct;
  
  strcpy(InFileName,FileName);
  sprintf(temp,".a%if",Dim);
  strcat(InFileName,temp);
  
  if(!(MyMultiStruct=(MultiStruct *) malloc(sizeof(MultiStruct)))) 
    Error( "Memory allocation problems (ReadMultiStruct)" );
  
  Print(_DNormal,"ReadMultiStruct: Reading `%s' \n",FileName);
  
  if(!(InFile=fopen(InFileName,"rb")))	
    Error("Error opening file: `%s'",InFileName);
  
  if(!(MyMultiStruct=(MultiStruct *) malloc(sizeof(MultiStruct)))) 
    Error( "Memory allocation problems (ReadMultiStruct)" );
  
  /* Read MultiStructHeader from file*/
  if((fscanf(InFile,"%s",temp))==EOF) Error("Cannot read file (ASCII)\n");
  if(strcmp(temp,"#MultiStruct"))
    Error("Error reading ASCII MultiStruct 1\n");
  
  if((fscanf(InFile,"%i",&MyMultiStruct->Id))==EOF) Error("Cannot read file (ASCII)\n");
  if(MyMultiStruct->Id!=20036) 
    Error("Error reading ASCII MultiStruct 2\n"); 
  
  if((fscanf(InFile,"%i",&MyMultiStruct->LengthOfHeader))==EOF) 
    Error("Cannot read file (ASCII)\n");
  if((fscanf(InFile,"%i",&MyMultiStruct->Dim))==EOF) 
    Error("Cannot read file (ASCII)\n");
  if((fscanf(InFile,"%i",&MyMultiStruct->NoOfSamples))==EOF) 
    Error("Cannot read file (ASCII)\n");
  if((fscanf(InFile,"%i",&MyMultiStruct->Type))==EOF) 
    Error("Cannot read file (ASCII)\n");
  if((fgets(MyMultiStruct->FileName,160,InFile))==NULL) 
    Error("Cannot read file (ASCII)\n");
  if((fgets(MyMultiStruct->Description,80,InFile))==NULL) 
    Error("Cannot read file (ASCII)\n");
  if((fgets(MyMultiStruct->Date,8,InFile))==NULL) 
    Error("Cannot read file (ASCII)\n");
  if((fscanf(InFile,"%i",&MyMultiStruct->ArrayType))==EOF) 
    Error("Cannot read file (ASCII)\n");

  if (MyMultiStruct->Dim!=Dim)
    Error("MultiStruct wrong dim. (Not supported)\n");

  MyMultiStruct->Samples=IntVector(Dim);
  MyMultiStruct->Min=FloatVector(Dim);
  MyMultiStruct->Delta=FloatVector(Dim);
   
  if((fscanf(InFile,"%s",temp))==EOF) 
    Error("Cannot read file (ASCII)\n");
  if(strcmp(temp,"#Dim"))
    Error("Error reading ASCII MultiStruct 3\n");
  for (m=0;m<Dim;m++) 
    fscanf(InFile,"%i",&(MyMultiStruct->Samples[m]));
    
  if((fscanf(InFile,"%s",temp))==EOF) 
    Error("Cannot read file (ASCII)\n");
  if(strcmp(temp,"#Min"))
    Error("Error reading ASCII MultiStruct 4\n");
  for (m=0;m<Dim;m++) 
    fscanf(InFile,"%e",&(MyMultiStruct->Min[m]));
    
  if((fscanf(InFile,"%s",temp))==EOF) 
    Error("Cannot read file (ASCII)\n");
  if(strcmp(temp,"#Delta"))
    Error("Error reading ASCII MultiStruct 5\n");
  for (m=0;m<Dim;m++) 
    fscanf(InFile,"%e",&(MyMultiStruct->Delta[m]));
 
  MyMultiStruct->Signal=NULL;
 
  if((fscanf(InFile,"%s",temp))==EOF) 
    Error("Cannot read file (ASCII)\n");
  if(strcmp(temp,"#Signal"))
    Error("Error reading ASCII MultiStruct 6\n");

  fclose(InFile);
  return MyMultiStruct;
}


/****************************************************************
[NAME]
ReadMult

[SYNOPSIS]
MultiStruct * ReadMult(char *FileName,int Type, int Signal)


[DESCRIPTION]

The function will read a {\tt MultiStruct} of type {\tt Type} with
filename {\tt FileName}. If {\tt Signal} equals {\tt \_NoSignal} only
the header is read.

[USAGE] 
{\tt MyMultiStruct=ReadMultiStruct("My3D",\_F3F,\_Signal);}

This will read {\tt My3D.f3f}.

[REVISION]
21/6-95 Peter Philipsen.
****************************************************************/
MultiStruct * ReadMult(char *FileName,int Type, int Signal)
{
  MultiStruct *ReturnMultiStruct;

  ReturnMultiStruct=NULL;

  if (Type==_F3F) {
    if(Signal==_Signal) 
      ReturnMultiStruct=ReadMultiStruct(FileName,3);
    else 
      ReturnMultiStruct= ReadMultiStruct_NoSignal(FileName,3);}
  else if (Type==_A3F) {
    if(Signal==_Signal) 
      ReturnMultiStruct= ReadMultiStruct_ASCII(FileName,3);
    else 
      ReturnMultiStruct= ReadMultiStruct_ASCII_NoSignal(FileName,3);}
  else if (Type==_F4F) {
    if (Signal==_Signal) 
      ReturnMultiStruct= ReadMultiStruct(FileName,4);
    else 
      ReturnMultiStruct= ReadMultiStruct_NoSignal(FileName,4);}
  else if (Type==_A4F) {
    if(Signal==_Signal) ReturnMultiStruct= ReadMultiStruct_ASCII(FileName,4);
    else 
      ReturnMultiStruct= ReadMultiStruct_ASCII_NoSignal(FileName,4);}
  else Error("Wrong Type (ReadMult)\n");
  
  return ReturnMultiStruct;
}


/****************************************************************
[NAME]
WriteMult

[SYNOPSIS]
MultiStruct * WriteMult(MultiStruct *MyMultiStruct,int Type, int Signal)


[DESCRIPTION]

The function will Write a {\tt MultiStruct} of type {\tt Type}. If
{\tt Signal} equals {\tt \_NoSignal} only the header is written.

[USAGE] 
{\tt MyMultiStruct=WriteMultiStruct(MyStruct,\_F3F,\_Signal);}

This will Write {\tt My3D.f3f}.

[REVISION]
21/6-95 Peter Philipsen.
****************************************************************/
void WriteMult(MultiStruct *MyMultiStruct,int Type, int Signal)
{
  if (Type==_F3F) {
    if(Signal==_Signal) 
      WriteMultiStruct(MyMultiStruct);
    else  
      WriteMultiStruct_NoSignal(MyMultiStruct);}
  else if (Type==_A3F) {
    if(Signal==_Signal)  
      WriteMultiStruct_ASCII(MyMultiStruct);
    else  
      WriteMultiStruct_ASCII_NoSignal(MyMultiStruct);}
  else if (Type==_F4F) {
    if(Signal==_Signal)  
      WriteMultiStruct(MyMultiStruct);
    else  
      WriteMultiStruct_NoSignal(MyMultiStruct);}
  else if (Type==_A4F) {
    if(Signal==_Signal)  
      WriteMultiStruct_ASCII(MyMultiStruct);
    else  
      WriteMultiStruct_ASCII_NoSignal(MyMultiStruct);}
  else 
    Error("Wrong Type (WriteMult)\n");
}


void MultiStruct_to_PPM(MultiStruct *My3D,
			char *BaseName,
			int Direction,
			int Oversamp,
			int Colormap,
                        float cor_alpha,
                        float cor_beta)
{
  int off,ii,iii,iiii,jj,kk,XSamples,YSamples,ZSamples,p1,p2,p3;
  int Samples1,Samples2,Samples3,iinterp1,iinterp2,res;
  float maxms,minms,alpha,beta,weight1,weight2;
  unsigned char Red_Colormap[256];
  unsigned char Green_Colormap[256];
  unsigned char Blue_Colormap[256];
  int *MyIntVolume;
  float *Signal;
  FILE *ppmfile;
  char OutFileName[200];
  
/*  Print(_DNormal,"Basename = %s\n",BaseName);*/
  
  if (My3D->Dim!=3)
    Error("The MultiStruct is not 3D (MultiStruct_to_PPM)");
  
  switch(Colormap)
  {
    case 0: /* Gray */
      for (ii=0;ii<256;ii++)
      {
        Red_Colormap[ii]=(unsigned char)ii;
        Green_Colormap[ii]=(unsigned char)ii;
        Blue_Colormap[ii]=(unsigned char)ii;
      }
      Print(_DNormal,"Colormap is Gray\n");
      break;
    case 1: /* red blue*/
      for (ii=0;ii<256;ii++)
      {
        Red_Colormap[ii]=(unsigned char)ii;
        Green_Colormap[ii]=(unsigned char)0;
        Blue_Colormap[ii]=(unsigned char)(255-ii);
      }
      Print(_DNormal,"Colormap is Red/Blue\n");
      break;
    case 2: /* rgb */
      for (ii=0;ii<256;ii++)
      {
        Red_Colormap[ii]=(unsigned char)(127-127*sin(M_PI*ii/255.0));
        Green_Colormap[ii]=(unsigned char)(127-127*sin(M_PI*ii/255.0-2*M_PI/6.0));
        Blue_Colormap[ii]=(unsigned char)(127-127*sin(M_PI*ii/255.0+M_PI/3.0));
      }
      Print(_DNormal,"Colormap is Red/Green/Blue\n");
      break;
    case 3: /* Pseudo rgb */
      for (ii=0;ii<256;ii++)
      {
        Red_Colormap[ii]=(unsigned char)ii;
        Green_Colormap[ii]=(unsigned char)(255-(128-ii)*(128-ii)/65);
        Blue_Colormap[ii]=(unsigned char)(255-ii);
      }
      Print(_DNormal,"Colormap is Pseudo Red/Green/Blue\n");
      break;
    default:
      Error("Wrong colormap selected");
  }

  XSamples=My3D->Samples[0];
  YSamples=My3D->Samples[1];
  ZSamples=My3D->Samples[2];
  
  MyIntVolume=IntVector(XSamples*YSamples*ZSamples);
  
  minms=MinMultiStruct(My3D);
  maxms=MaxMultiStruct(My3D);
  
  beta=0.1-minms*255.9/(maxms-minms);
  alpha=255.9/(maxms-minms);
  
  Signal=My3D->Signal;
  for (ii=0;ii<My3D->NoOfSamples;ii++)
  {
    res=(int)(cor_alpha*alpha*Signal[ii]+beta+cor_beta);
    if (res<0) res=0;
    if (res>255) res=255;
    MyIntVolume[ii]=res;
  }

  Samples1=Samples2=Samples3=p1=p2=p3=1;
  switch(Direction)
  {
    case 1:
      p1=1;
      p2=XSamples;
      p3=XSamples*YSamples;
      Samples1=XSamples;
      Samples2=YSamples;
      Samples3=ZSamples;
      Print(_DNormal,"Images is (Y,Z). The number indicates the X-direction\n");
      break;
    case 2:
      p1=XSamples;
      p2=1;
      p3=XSamples*YSamples;
      Samples1=YSamples;
      Samples2=XSamples;
      Samples3=ZSamples;
      Print(_DNormal,"Images is (X,Z). The number indicates the Y-direction\n");
      break;
    case 3:
      p1=XSamples*YSamples;
      p2=1;
      p3=XSamples;
      Samples1=ZSamples;
      Samples2=XSamples;
      Samples3=YSamples;
      Print(_DNormal,"Images is (X,Y). The number indicates the Z-direction\n");
      break;
    default:
      Error("Wrong direction");
  }
  
 
  for (ii=0;ii<Samples1;ii++)
  {
    for (iii=0;iii<Oversamp;iii++)
    {
      iiii=ii*Oversamp+iii;
      Print(_DNormal,"Doing image %i (%i)\r",iiii,Samples1*Oversamp);
      sprintf(OutFileName,"%s_%03i.ppm",BaseName,iiii);
      ppmfile=fopen(OutFileName,"wt");
      fprintf(ppmfile,"P6\n%d %d\n255\n",Samples3,Samples2);
      iinterp2=iinterp1=ii; 
      if (ii<Samples1-1) iinterp2++;
      weight1=(Oversamp-iii)/(float)Oversamp;
      weight2=iii/(float)Oversamp;
            

      for (jj=Samples2-1;jj>=0;jj--)	
        for (kk=0;kk<Samples3;kk++)
        {
	  off=jj*p2+kk*p3;
	  res=(int)(weight1*MyIntVolume[iinterp1*p1+off]
		    +weight2*MyIntVolume[iinterp2*p1+off]);
          
	  fwrite(&Red_Colormap[res],1,1,ppmfile);
	  fwrite(&Green_Colormap[res],1,1,ppmfile);
	  fwrite(&Blue_Colormap[res],1,1,ppmfile);
	}
      fclose(ppmfile);
    }
  }
  Print(_DNormal,"Done writing images  %s_[%03i-%03i].ppm       \n",
        BaseName,0,Samples1*Oversamp);
}

