/***************************************************************************
[HEADER]
  
This is the central program, which contains routines to call the main
calculation routines.
  
***************************************************************************/
#include <math.h>
#include <time.h>
#include <stdio.h>
#include <ctype.h>
#include <string.h>
#include <stdarg.h>
#include <malloc.h>
#include "Recon3D.h"

char *IniBuffer;


/****************************************************************
[NAME]
CSreconstruct3D

[SYNOPSIS]
void CSreconstruct3D(void)

[DESCRIPTION]
The function implements the Fourier Slice Theorem

[USAGE]
{\tt CSreconstruct()}

[REVISION]
14/11-95 Peter Toft
****************************************************************/
void CSreconstruct3D(void)
{
  MultiStruct *StructRadon, *StructVol;
  XYZDef *MyXYZ;
  TrigDef *MyTrig;
  float Tid1,Tid2;
 
  StructRadon=ReadMult(IniFile3D.Sinogram,IniFile3D.SinogramType,_Signal);
  InfoMultiStruct(_DNormal,StructRadon);

  MyXYZ=GenerateXYZ_fromIniFile();

  MyTrig=GenerateTrigDef_fromMultiStruct(StructRadon);
  Tid1=clock();

  if (IniFile3D.Select_Par_1==0)
    StructVol=CSrebin3D_NN(StructRadon,MyTrig,MyXYZ);
  else
    StructVol=CSrebin3D_Tl(StructRadon,MyTrig,MyXYZ);

  Tid2=clock();

  Print(_DNormal,"(CSreconstruct3D) used %f sec\n",
        (Tid2-Tid1)/(float)CLOCKS_PER_SEC);

  RenameMultiStruct(StructVol,IniFile3D.Volume);
  InfoMultiStruct(_DNormal,StructVol);
  
  WriteMult(StructVol,IniFile3D.VolumeType,_Signal);

  FreeMultiStruct(StructRadon);
  FreeMultiStruct(StructVol);
}


/****************************************************************
[NAME]
FABreconstruct3D

[SYNOPSIS]
void FABreconstruct3D(void)

[DESCRIPTION]
The function implements filtering after backprojection.

[USAGE]
{\tt FABreconstruct()}

[REVISION]
20/6-95 Peter and Peter.\\
6/9-95 Peter Toft - Transferring limitation angle psi to 
filtervolume.
****************************************************************/
void FABreconstruct3D(void)
{
  MultiStruct *StructRadon, *StructVol,*StructVol3;
  XYZDef *OrgXYZ,*MyXYZ;
  TrigDef *MyTrig;
  int i;
  int *Radix2Samples,*Position,*Samples;
  float *NewMin;

  Radix2Samples=IntVector(3);
  Samples=IntVector(3);
  Position=IntVector(3);
  NewMin=FloatVector(3);

  StructRadon=ReadMult(IniFile3D.Sinogram,IniFile3D.SinogramType,_Signal);
  InfoMultiStruct(_DNormal,StructRadon);

  MyTrig=GenerateTrigDef_fromMultiStruct(StructRadon);
  
  OrgXYZ=GenerateXYZ_fromIniFile();

  Radix2Samples[0]=NextRadix2(OrgXYZ->XSamples);
  Radix2Samples[1]=NextRadix2(OrgXYZ->YSamples);
  Radix2Samples[2]=NextRadix2(OrgXYZ->ZSamples);
 
  NewMin[0]=OrgXYZ->Xmin-
    OrgXYZ->DeltaX*(Radix2Samples[0]-OrgXYZ->XSamples+1)/2;
  NewMin[1]=OrgXYZ->Ymin-
    OrgXYZ->DeltaY*(Radix2Samples[1]-OrgXYZ->YSamples+1)/2;
  NewMin[2]=OrgXYZ->Zmin-
    OrgXYZ->DeltaZ*(Radix2Samples[2]-OrgXYZ->ZSamples+1)/2;
 

  MyXYZ=GenerateXYZ(Radix2Samples[0],Radix2Samples[1],Radix2Samples[2],
		    NewMin[0],NewMin[1],NewMin[2],
		    OrgXYZ->DeltaX,OrgXYZ->DeltaY,OrgXYZ->DeltaZ);

  
  StructVol=BackProject3D(StructRadon,MyXYZ,MyTrig);

  Real2ComplexMultiStruct(StructVol);
  FilterVolume(StructVol,-MyTrig->Phimin);
  RealMultiStruct(StructVol);

  for (i=0;i<3;i++)
     Position[i]=0;
 
  Samples[0]=OrgXYZ->XSamples;
  Samples[1]=OrgXYZ->YSamples;
  Samples[2]=OrgXYZ->ZSamples;
 
  StructVol3=ResizeMultiStruct(StructVol,Samples,Position);

  RenameMultiStruct(StructVol3,IniFile3D.Volume);

  InfoMultiStruct(_DNormal,StructVol3);
  WriteMult(StructVol3,IniFile3D.VolumeType,_Signal);
}


/****************************************************************
[NAME]
FBreconstruct3D

[SYNOPSIS]
void FBreconstruct3D(void)

[DESCRIPTION]
The function implements filtered backprojection.

[USAGE]
{\tt FBreconstruct()}

[REVISION]
20/6-95 Peter and Peter.
****************************************************************/
void FBreconstruct3D(void)
{
  MultiStruct *StructRadon, *StructVol;
  XYZDef *MyXYZ;
  TrigDef *MyTrig;
  float Tid1,Tid2,Tid3;
 
  StructRadon=ReadMult(IniFile3D.Sinogram,IniFile3D.SinogramType,_Signal);
  InfoMultiStruct(_DNormal,StructRadon);

  MyXYZ=GenerateXYZ_fromIniFile();

  MyTrig=GenerateTrigDef_fromMultiStruct(StructRadon);
  Tid1=clock();

  /* Set Select_Par_1=1 if you want to window the sinogram, and set 
     Direct_Par_1 to a value greater than zero to use bending the ramp
     filter. This parameter is relative to the maximum frequency. E.g. a 
     value of 0.5 or less will suppress the high frequencies. */

  Filter4DSinogram(StructRadon,
                   MyTrig,
                   -MyTrig->Phimin,
                   IniFile3D.Select_Par_1,
                   IniFile3D.Direct_Par_1);

  Tid2=clock();

  StructVol=BackProject3D(StructRadon,MyXYZ,MyTrig);

  Tid3=clock();

  Print(_DNormal,"(FBreconstruct3D) Filtering used %f sec\n",
        (Tid2-Tid1)/(float)CLOCKS_PER_SEC);
  Print(_DNormal,"(FBreconstruct3D) Backprojection used %f sec\n",
        (Tid3-Tid2)/(float)CLOCKS_PER_SEC);
  RenameMultiStruct(StructVol,IniFile3D.Volume);
  InfoMultiStruct(_DNormal,StructVol);
  
  WriteMult(StructVol,IniFile3D.VolumeType,_Signal);

  FreeMultiStruct(StructRadon);
  FreeMultiStruct(StructVol);
}


/****************************************************************
[NAME]
ART3D

[SYNOPSIS]
void ART3D(void)

[DESCRIPTION]
The function implements 3D ART.

[USAGE]
{\tt ART3D()}

[REVISION]
28/6-95 Peter Toft.
****************************************************************/
void ART3D(void)
{
  MultiStruct *StructRadon, *StructVol;
  XYZDef *MyXYZ;

  StructRadon=ReadMult(IniFile3D.Sinogram,IniFile3D.SinogramType,_Signal);
  InfoMultiStruct(_DNormal,StructRadon);

  MyXYZ=GenerateXYZ_fromIniFile();

  StructVol=Iterative_ART_3D(StructRadon,MyXYZ,IniFile3D.Iterations);
  
  RenameMultiStruct(StructVol,IniFile3D.Volume);
  InfoMultiStruct(_DNormal,StructVol);
  
  WriteMult(StructVol,IniFile3D.VolumeType,_Signal);

  FreeMultiStruct(StructRadon);
  FreeMultiStruct(StructVol);
}


/****************************************************************
[NAME]
MART3D

[SYNOPSIS]
void MART3D(void)

[DESCRIPTION]
The function implements 3D MART.

[USAGE]
{\tt MART3D()}

[REVISION]
19/9-95 Peter Toft.
****************************************************************/
void MART3D(void)
{
  MultiStruct *StructRadon, *StructVol;
  XYZDef *MyXYZ;

  StructRadon=ReadMult(IniFile3D.Sinogram,IniFile3D.SinogramType,_Signal);
  InfoMultiStruct(_DNormal,StructRadon);

  MyXYZ=GenerateXYZ_fromIniFile();

  StructVol=Iterative_MART_3D(StructRadon,MyXYZ,IniFile3D.Iterations);
  
  RenameMultiStruct(StructVol,IniFile3D.Volume);
  InfoMultiStruct(_DNormal,StructVol);
  
  WriteMult(StructVol,IniFile3D.VolumeType,_Signal);

  FreeMultiStruct(StructRadon);
  FreeMultiStruct(StructVol);
}


/****************************************************************
[NAME]
EM3D

[SYNOPSIS]
void EM3D(void)

[DESCRIPTION]
The function implements 3D EM.

[USAGE]
{\tt EM3D()}

[REVISION]
24/8-95 Peter Toft.
****************************************************************/
void EM3D(void)
{
  MultiStruct *StructRadon, *StructVol;
  XYZDef *MyXYZ;

  StructRadon=ReadMult(IniFile3D.Sinogram,IniFile3D.SinogramType,_Signal);
  InfoMultiStruct(_DNormal,StructRadon);

  MyXYZ=GenerateXYZ_fromIniFile();

  StructVol=Iterative_EM_3D(StructRadon,MyXYZ,IniFile3D.Iterations);
  
  RenameMultiStruct(StructVol,IniFile3D.Volume);
  InfoMultiStruct(_DNormal,StructVol);
  
  WriteMult(StructVol,IniFile3D.VolumeType,_Signal);

  FreeMultiStruct(StructRadon);
  FreeMultiStruct(StructVol);
}


/****************************************************************
[NAME]
Forward3D

[SYNOPSIS]
void Forward3D(void)

[DESCRIPTION]
The function implements a numerical calculation of the line 
integrals.

[USAGE]
{\tt Forward3D()}

[NOTE]
Uses the {\tt OrgFile} to specify the 4D Sinogram.

[REVISION]
20/6-95 Peter and Peter.
****************************************************************/
void Forward3D(void)
{
  MultiStruct *StructRadon2, *StructRadon, *StructVol;
  TrigDef *MyTrig;

  if ((IniFile3D.OrgFileType!=_F4F) && (IniFile3D.OrgFileType!=_A4F))
    Error("OrgFile must be a 4 dimensional MultiStruct\n");

  StructRadon2=ReadMult(IniFile3D.OrgFile,IniFile3D.OrgFileType,_NoSignal); 

  InfoMultiStruct(_DNormal,StructRadon2);
  MyTrig=GenerateTrigDef_fromMultiStruct(StructRadon2);
  StructVol=ReadMult(IniFile3D.Volume,IniFile3D.VolumeType,_Signal);
  StructRadon=NumericalRadon3D(StructVol,MyTrig);
  RenameMultiStruct(StructRadon,IniFile3D.Sinogram);
  InfoMultiStruct(_DNormal,StructRadon);
  WriteMult(StructRadon,IniFile3D.SinogramType,_Signal);

  FreeMultiStruct(StructRadon2);
  FreeMultiStruct(StructRadon);
  FreeMultiStruct(StructVol);
}

void Test(void)
{
  MultiStruct *StructVol;
  int Samples[3];
  int x,y,z;

  Samples[0]=Samples[1]=Samples[2]=20;
  StructVol=NewMultiStruct("Hej",3,Samples,_RealArray);

  for (x=0;x<20;x++)
    StructVol->Signal[x]=1;
  for (y=0;y<20;y++)
    StructVol->Signal[20*y]=1;
  for (z=0;z<20;z++)
    StructVol->Signal[20*z*20]=1;

  InfoMultiStruct(_DNormal,StructVol);
  BoxFilterVolume(StructVol,2);
  InfoMultiStruct(_DNormal,StructVol);
  WriteMult(StructVol,_F3F,_Signal);
  FreeMultiStruct(StructVol);
}


/****************************************************************
[NAME]
Convert

[SYNOPSIS]
void Convert(void)

[DESCRIPTION]
The function converts the Volume and Sinogram from ASCII to binary 
or opposite depending on the types. The function writes in the same 
filenames with opposite extentions. 

[USAGE]
{\tt Convert()}

[REVISION]
20/6-95 Peter and Peter.
****************************************************************/
void Convert(void)
{
  MultiStruct *StructRadon, *StructVol;
  
  StructRadon=ReadMult(IniFile3D.Sinogram,IniFile3D.SinogramType,_Signal);
  RenameMultiStruct(StructRadon,IniFile3D.Sinogram);
  if (IniFile3D.SinogramType==_F4F)
    WriteMult(StructRadon,_A4F,_Signal);
  else
    WriteMult(StructRadon,_F4F,_Signal);

  StructVol=ReadMult(IniFile3D.Volume,IniFile3D.VolumeType,_Signal);
  RenameMultiStruct(StructVol,IniFile3D.Volume);
  if (IniFile3D.VolumeType==_F4F)
    WriteMult(StructVol,_A3F,_Signal);
  else
    WriteMult(StructVol,_F3F,_Signal);

  FreeMultiStruct(StructRadon);
}

/****************************************************************
[NAME]
ConvertFIF

[SYNOPSIS]
void ConvertFIF(void)

[DESCRIPTION]
The function converts the Volume as individual FIF files to a 
single F3F file. 

[USAGE]
{\tt ConvertFIF()}

[REVISION]
20/6-95 Peter and Peter.\\
May 97 Peter Toft Missing arguments - disabling the function.
****************************************************************/
void ConvertFIF(void)
{ 
  /*
  MultiStruct *StructVol;
  StructVol=FIFs_to_F3F();
  RenameMultiStruct(StructVol,IniFile3D.Volume);
  WriteMult(StructVol,_A3F,_Signal);

  FreeMultiStruct(StructRadon);*/
}


/***************************************************************************
[NAME]
Main

[SYNOPSIS]
it IniFile.ini

[DESCRIPTION] 

This is the main function. It reads the specified {\tt IniFile.ini}
file and executes one of the following functions:

\begin{itemize}
\item Central Slice (CS) experimental!
\item Filtered BackProjection (FB).
\item Filtering after BackProjection (FAB).
\item Algebraic Reconstruction Technique (ART).
\item Multiplicative Algebraic Reconstruction Technique (MART).
\item Expectation Maximization (EM).
\item Convert between ASCII and Binary (Convert).
\item Numerical Radon Transform (RT).
\end{itemize}

[USAGE] 

{\tt it Test.ini} 

Executes the program and uses {\tt Test.ini} as input.

[REVISION]
Jun. 95, PAP\\
Dec 13 95 Peter Toft\\
April 13, 96, Peter Toft
***************************************************************************/
void main(int argc,char * argv[])
{
  int RealTid1,RealTid2;
  float Tid;
  char Value[100];
  
  GetDateTime(Value,_RealTime);
  sscanf(Value,"%i",&RealTid1);
  Tid=clock();
  
  if(!(argc==2))
    Error("3Dradon, PToft&PAP 1995: You must specify INI file");
  IniBuffer=ReadIni(argv[1]);

  Read3DArgs(IniBuffer);

  Print(_DNormal,"\n**************************************\n");
  Print(_DNormal,"* 3Dradon, rev 1.10 PToft & PAP 1995 *\n");
  Print(_DNormal,"**************************************\n");

  GetArg(IniBuffer,"Function",Value);
  if (strstr(Value,"FB")) 
    FBreconstruct3D();
  else if (strstr(Value,"FAB")) 
    FABreconstruct3D();
  else if (strstr(Value,"CS")) 
    CSreconstruct3D();
  else if (strstr(Value,"Forward")) 
    Forward3D();
  else if (strstr(Value,"Convert")) 
    Convert();
  else if (strstr(Value,"MART")) 
    MART3D();
  else if (strstr(Value,"ART")) 
    ART3D();
  else if (strstr(Value,"EM")) 
    EM3D();
  else if (strstr(Value,"Test")) 
    Test();
  else 
    Error("Function not recognized: '%s'",Value);

  GetDateTime(Value,_RealTime);
  sscanf(Value,"%i",&RealTid2);
  Tid=(clock()-Tid)/(float)CLOCKS_PER_SEC;
  Print(_DNormal,"3Dradon:   Program was active for %.2f seconds\n",Tid) ;
  Print(_DNormal,"           World time elapsed %d seconds\n",
	(RealTid2-RealTid1));
  Print(_DNormal,"           Program used %.2f %% cpu time\n",
         (Tid>((float)RealTid2-RealTid1))?99.99:(Tid/((float)RealTid2-RealTid1)*100));
  Print(_DNormal,"-----------------------\n");
  CloseLog();
}












