/************************************************************************/
/*                                                                      */
/*  vol_view is a program to show f3f files usign the xforms toolkit.   */
/*                                                                      */
/*                                                                      */
/* Usage :                                                              */
/* vol_view multistruct-filename zoomx zoomy zoomz filter-half-width    */
/*                                                                      */
/*                                                                      */
/*  vol_view is written by Peter Toft.                                  */
/*                                                                      */
/************************************************************************/


#include <stdlib.h>
#include <math.h>
#include <string.h>
#include "forms.h"
#include "vol.h"
#include "Recon3D.h"


FL_OBJECT *pixxy,*slix,*pixyz;
FL_OBJECT *sliy,*pixxz,*sliz,*slicolmin,*slicolmax;
FL_OBJECT *check1,*check2,*check3;
FL_OBJECT *slixyx,*slixyy,*sliyzy,*sliyzz,*slixzx,*slixzz;

#define maxzoomfactor 10

char* myvol;
char **myxyimage,**myyzimage,**myxzimage;
int X,Y,Z;
char **mygraycolormap,**myhotcolormap,**myrgbcolormap;
int choisecolormap,zoomfactorx,zoomfactory,zoomfactorz;
int x,y,z,mincol,maxcol;


char hexc(int ii)
{
  if (ii<10)
    return ii+48;
  else
    return ii+87;
}

void makegraycolormap(void)
{
  int kk,ii,jj,intcol,lo,up;
  float slope;
  Print(_DDebug,"(makecolormaps)\n");

  slope=255/(float)(maxcol-mincol);

printf("hej\n");
  /* Gray */
  kk=0;
  for (ii=0;ii<16;ii++)
    for (jj=0;jj<16;jj++) {
      mygraycolormap[kk]=(char *)malloc(13);
      mygraycolormap[kk][0]=65+ii;
      mygraycolormap[kk][1]=65+jj;
      mygraycolormap[kk][2]=' ';
      mygraycolormap[kk][3]='c';
      mygraycolormap[kk][4]=' ';
      mygraycolormap[kk][5]='#';
      intcol=(int)(slope*(kk-mincol));
      if (intcol>255) intcol=255;
      if (intcol<0) intcol=0;
      up=(int)(intcol/16);
      lo=intcol%16;
      mygraycolormap[kk][10]=mygraycolormap[kk][8]=mygraycolormap[kk][6]=hexc(up);
      mygraycolormap[kk][11]=mygraycolormap[kk][9]=mygraycolormap[kk][7]=hexc(lo);
      mygraycolormap[kk][12]='\0';
      kk++;
    }
}

void makehotcolormap(void)
{
  int kk,ii,jj,r,g,b,intcol;
  float slope;
  Print(_DDebug,"(makecolormaps)\n");

  slope=255/(float)(maxcol-mincol);

  /* Hot Metal*/
  kk=0;
  for (ii=0;ii<16;ii++)
    for (jj=0;jj<16;jj++) {
      myhotcolormap[kk]=(char *)malloc(13);
      myhotcolormap[kk][0]=65+ii;
      myhotcolormap[kk][1]=65+jj;
      myhotcolormap[kk][2]=' ';
      myhotcolormap[kk][3]='c';
      myhotcolormap[kk][4]=' ';
      myhotcolormap[kk][5]='#';
      intcol=(int)(slope*(kk-mincol));
      if (intcol>255) intcol=255;
      if (intcol<0) intcol=0;
      r=(int)(intcol*2.55);
      if (r>255) 
        r=255;
      g=(int)((intcol-1e2)*2.9);
      if (g<0)
        g=0;
      if (g>255) 
        g=255;
      b=(int)((intcol-1.92e2)*4.05);
      if (b<0)
        b=0;
      if (b>255) 
        b=255;
      myhotcolormap[kk][6]=hexc((int)(r/16));
      myhotcolormap[kk][7]=hexc(r%16);
      myhotcolormap[kk][8]=hexc((int)(g/16));
      myhotcolormap[kk][9]=hexc(g%16);
      myhotcolormap[kk][10]=hexc((int)(b/16));
      myhotcolormap[kk][11]=hexc(b%16);
      myhotcolormap[kk][12]='\0';
      kk++;
    }
}

void makergbcolormap(void)
{
  int kk,ii,jj,r,g,b,intcol;
  float slope;
  Print(_DDebug,"(makecolormaps)\n");

  slope=255/(float)(maxcol-mincol);

  /* Rgb */
  kk=0;
  for (ii=0;ii<16;ii++)
    for (jj=0;jj<16;jj++) {
      myrgbcolormap[kk]=(char *)malloc(13);
      myrgbcolormap[kk][0]=65+ii;
      myrgbcolormap[kk][1]=65+jj;
      myrgbcolormap[kk][2]=' ';
      myrgbcolormap[kk][3]='c';
      myrgbcolormap[kk][4]=' ';
      myrgbcolormap[kk][5]='#';
      intcol=(int)(slope*(kk-mincol));
      if (intcol>255) intcol=255;
      if (intcol<0) intcol=0;
      r=(int)(255*0.5*(1+sin(2*intcol*M_PI/255)));
      g=(int)(255*0.5*(1+sin(2*(intcol-256/3.0)*M_PI/255)));
      b=(int)(255*0.5*(1+sin(2*(intcol+256/3.0)*M_PI/255)));
      myrgbcolormap[kk][8]=hexc((int)(r/16));
      myrgbcolormap[kk][9]=hexc(r%16);
      myrgbcolormap[kk][10]=hexc((int)(g/16));
      myrgbcolormap[kk][11]=hexc(g%16);
      myrgbcolormap[kk][6]=hexc((int)(b/16));
      myrgbcolormap[kk][7]=hexc(b%16);
      myrgbcolormap[kk][12]='\0';
      kk++;
    }
}


void check_cb(FL_OBJECT *ob, long data)
{
  int ii;

  Print(_DNormal,"colormap = %l (check_cb)\n",data);
  if (data==1) 
    for (ii=0;ii<256;ii++)
    {
      myxyimage[ii+1]=mygraycolormap[ii];
      myxzimage[ii+1]=mygraycolormap[ii];
      myyzimage[ii+1]=mygraycolormap[ii];
    }
  if (data==2) 
    for (ii=0;ii<256;ii++)
    {
      myxyimage[ii+1]=myhotcolormap[ii];
      myxzimage[ii+1]=myhotcolormap[ii];
      myyzimage[ii+1]=myhotcolormap[ii];
    }
  if (data==3) 
    for (ii=0;ii<256;ii++)
    {
      myxyimage[ii+1]=myrgbcolormap[ii];
      myxzimage[ii+1]=myrgbcolormap[ii];
      myyzimage[ii+1]=myrgbcolormap[ii];
    }
  fl_set_pixmap_data(pixxy,myxyimage);
  fl_set_pixmap_data(pixxz,myxzimage);
  fl_set_pixmap_data(pixyz,myyzimage);
  fl_redraw_object(pixxy);
  fl_redraw_object(pixxz);
  fl_redraw_object(pixyz);
}

void mincol_cb(FL_OBJECT *ob, long data)
{
  mincol=fl_get_slider_value(ob);
  if (mincol>maxcol) 
  {
    maxcol=mincol;
    fl_set_slider_value(slicolmax,maxcol);
  }
  if (choisecolormap==1)
    makegraycolormap();
  else
    if (choisecolormap==2)
      makehotcolormap();
    else
      makergbcolormap();

  check_cb(check1,(long)choisecolormap);
}

void maxcol_cb(FL_OBJECT *ob, long data)
{
  maxcol=fl_get_slider_value(ob);
  if (mincol>maxcol) 
  {
    mincol=maxcol;
    fl_set_slider_value(slicolmin,mincol);
  }
  check_cb(check1,(long)choisecolormap);
}

/* callbacks for form vol */
void mycountxy(FL_OBJECT *ob, long data)
{
  /* fill-in code for callback */
  int ii,jj,kk,zz,off,off2,off3;
  char cu,cl;
  char str[30];
  char *imptr;

  z=kk=fl_get_slider_value(sliz);
  Print(_DDebug,"z = %i (mycountxy)\n",z);

  for (jj=0;jj<Y;jj++) 
  {
    off3=(Y-1-jj)*zoomfactory+257;
    imptr=myxyimage[off3];
    for (ii=0;ii<X;ii++) 
    {
      off = 2*(kk*Y*X+jj*X+ii);
      cu=myvol[off];
      cl=myvol[off+1];
      for (zz=0;zz<zoomfactorx;zz++) 
      {
        off2=2*(ii*zoomfactorx+zz);
        imptr[off2]=cu;
        imptr[off2+1]=cl;
      }
    }
    for (zz=1;zz<zoomfactory;zz++) 
      memcpy(myxyimage[off3+zz],imptr,2*X*zoomfactorx);
  }

  fl_set_slider_value(slixzz,Z-1-z);
  fl_set_slider_value(sliyzz,Z-1-z);
  fl_set_pixmap_data(pixxy,myxyimage);
  fl_redraw_object(pixxy);
  sprintf(str,"Pos (%i,%i,%i)",x,y,z);
  fl_set_object_label((FL_OBJECT *)data,str);
}


/* callbacks for form vol */
void mycountyz(FL_OBJECT *ob, long data)
{
  /* fill-in code for callback */
  int ii,jj,kk,zz,off,off2,off3;
  char cu,cl;
  char str[30];
  char *imptr;

  x=kk=fl_get_slider_value(slix);
  Print(_DDebug,"x = %i (mycountyz)\n",x);

  for (jj=0;jj<Z;jj++) 
  {
    off3=(Z-1-jj)*zoomfactorz+257;
    imptr=myyzimage[off3];
    for (ii=0;ii<Y;ii++) 
    {
      off = 2*(jj*Y*X+ii*X+kk);
      cu=myvol[off];
      cl=myvol[off+1];
      for (zz=0;zz<zoomfactory;zz++) 
      {
        off2=2*(ii*zoomfactory+zz);
        imptr[off2]=cu;
        imptr[off2+1]=cl;
      }
    }
    for (zz=1;zz<zoomfactorz;zz++) 
      memcpy(myyzimage[off3+zz],imptr,2*Y*zoomfactory);
  }

  fl_set_slider_value(slixyx,x);
  fl_set_slider_value(slixzx,x);
  fl_set_pixmap_data(pixyz,myyzimage);
  fl_redraw_object(pixyz);
  sprintf(str,"Pos (%i,%i,%i)",x,y,z);
  fl_set_object_label((FL_OBJECT *)data,str);
}

/* callbacks for form vol */
void mycountxz(FL_OBJECT *ob, long data)
{
  /* fill-in code for callback */
  int ii,jj,kk,zz,off,off2,off3;
  char cu,cl;
  char str[30];
  char *imptr;

  y=kk=fl_get_slider_value(sliy);
  Print(_DDebug,"y = %i (mycountxz)\n",y);

  for (jj=0;jj<Z;jj++) 
  {
    off3=(Z-1-jj)*zoomfactorz+257;
    imptr=myxzimage[off3];
    for (ii=0;ii<X;ii++) 
    {
      off = 2*(jj*Y*X+kk*X+ii);
      cu=myvol[off];
      cl=myvol[off+1];
      for (zz=0;zz<zoomfactorx;zz++) 
      {
        off2=2*(ii*zoomfactorx+zz);
        imptr[off2]=cu;
        imptr[off2+1]=cl;
      }
    }
    for (zz=1;zz<zoomfactorz;zz++) 
      memcpy(myxzimage[off3+zz],imptr,2*X*zoomfactorx);
  }

  fl_set_slider_value(slixyy,Y-1-y);
  fl_set_slider_value(sliyzy,y);
  fl_set_pixmap_data(pixxz,myxzimage);
  fl_redraw_object(pixxz);
  sprintf(str,"Pos (%i,%i,%i)",x,y,z);
  fl_set_object_label((FL_OBJECT *)data,str);
}

FD_vol *create_form_vol(void)
{
  FL_OBJECT *obj,*xyzvalue,*zoomvalue;
  FD_vol *fdui = (FD_vol *) fl_calloc(1, sizeof(*fdui));
  char str[30];
  int maxdown,XX,YY,ZZ;

  Print(_DDebug,"(create_form_vol)\n");
  XX=X*zoomfactorx;
  YY=Y*zoomfactory;
  ZZ=Z*zoomfactorz;

  maxdown=YY;
  if (ZZ>maxdown) maxdown=ZZ;
  maxdown+=75;

  fdui->vol = fl_bgn_form(FL_NO_BOX,180+XX*2+YY, 150+maxdown);
  obj = fl_add_box(FL_UP_BOX,0,0,180+XX*2+YY, 150+maxdown,"");
  obj = fl_add_button(FL_RETURN_BUTTON,10, maxdown,60,30,"Exit");    

  pixxy = fl_add_pixmap(FL_NORMAL_PIXMAP,110,50,XX,YY,"");
  pixxz = fl_add_pixmap(FL_NORMAL_PIXMAP,130+XX,50,XX,ZZ,"");
  pixyz = fl_add_pixmap(FL_NORMAL_PIXMAP,150+2*XX,50,YY,ZZ,"");

  /*Slider for the xy-pixmap*/
  slixyx = fl_add_slider(FL_HOR_SLIDER,110,52+YY,XX,8,"Slice xy");
  fl_set_slider_bounds(slixyx,0,X-1);
  fl_set_slider_step(slixyx,1);

  slixyy = fl_add_slider(FL_VERT_SLIDER,100,50,8,YY,"");
  fl_set_slider_bounds(slixyy,0,Y-1);
  fl_set_slider_step(slixyy,1);

  /*Slider for the xz-pixmap*/
  slixzx = fl_add_slider(FL_HOR_SLIDER,130+XX,52+ZZ,XX,8,"Slice xz");
  fl_set_slider_bounds(slixzx,0,X-1);
  fl_set_slider_step(slixzx,1);

  slixzz = fl_add_slider(FL_VERT_SLIDER,120+XX,50,8,ZZ,"");
  fl_set_slider_bounds(slixzz,0,Z-1);
  fl_set_slider_step(slixzz,1);

  /*Slider for the yz-pixmap*/
  sliyzy = fl_add_slider(FL_HOR_SLIDER,150+2*XX,52+ZZ,YY,8,"Slice yz");
  fl_set_slider_bounds(sliyzy,0,Y-1);
  fl_set_slider_step(sliyzy,1);

  sliyzz = fl_add_slider(FL_VERT_SLIDER,140+2*XX,50,8,ZZ,"");
  fl_set_slider_bounds(sliyzz,0,Z-1);
  fl_set_slider_step(sliyzz,1);

  xyzvalue = fl_add_box(FL_DOWN_BOX,80+(int)(1.5*XX),10,100,30,"Slice y");  
  
  sliz = fl_add_slider(FL_HOR_SLIDER,105,maxdown,10+XX,30,"Select z-slice");
  sliy = fl_add_slider(FL_HOR_SLIDER,125+XX,maxdown,10+XX,30,"Select y-slice");
  slix = fl_add_slider(FL_HOR_SLIDER,145+2*XX,maxdown,10+YY,30,"Select x-slice");
  fl_set_slider_step(slix,1);
  fl_set_slider_value(slix,0);
  fl_set_slider_bounds(slix,0,X-1);
  fl_set_slider_precision(slix,0);
  fl_set_slider_return(slix,1);
  fl_set_object_callback(slix,mycountyz,(long)xyzvalue);

  fl_set_slider_step(sliy,1);
  fl_set_slider_value(sliy,0);
  fl_set_slider_bounds(sliy,0,Y-1);
  fl_set_slider_precision(sliy,0);
  fl_set_slider_return(sliy,1);
  fl_set_object_callback(sliy,mycountxz,(long)xyzvalue);

  fl_set_slider_step(sliz,1);
  fl_set_slider_value(sliz,0);
  fl_set_slider_bounds(sliz,0,Z-1);
  fl_set_slider_precision(sliz,0);
  fl_set_slider_return(sliz,1);
  fl_set_object_callback(sliz,mycountxy,(long)xyzvalue);
  
  check1 = fl_add_checkbutton(FL_RADIO_BUTTON,20,10,20,20,"Gray");
  check2 = fl_add_checkbutton(FL_RADIO_BUTTON,20,40,20,20,"Hot");
  check3 = fl_add_checkbutton(FL_RADIO_BUTTON,20,70,20,20,"Rgb");

  fl_set_button(check1,1);
  fl_set_object_callback(check1,check_cb,(long)1);
  fl_set_object_callback(check2,check_cb,(long)2);
  fl_set_object_callback(check3,check_cb,(long)3);

  zoomvalue = fl_add_box(FL_DOWN_BOX,10,100,80,30,"zoomfactor");  
  sprintf(str,"zoom=(%i,%i,%i)",zoomfactorx,zoomfactory,zoomfactorz);
  fl_set_object_label(zoomvalue,str);

  slicolmin = fl_add_slider(FL_HOR_SLIDER,100,maxdown+50,100,20,"Colormin");
  slicolmax = fl_add_slider(FL_HOR_SLIDER,210,maxdown+50,100,20,"Colormax");
  fl_set_slider_step(slicolmin,1);
  fl_set_slider_step(slicolmax,1);
  fl_set_slider_value(slicolmin,0);
  fl_set_slider_value(slicolmax,255);
  fl_set_slider_bounds(slicolmin,0,255);
  fl_set_slider_bounds(slicolmax,0,255);
  fl_set_slider_precision(slicolmin,1);
  fl_set_slider_precision(slicolmax,1);
  fl_set_slider_return(slicolmin,1);
  fl_set_slider_return(slicolmax,1);
  fl_set_object_callback(slicolmin,mincol_cb,(long)slicolmax);
  fl_set_object_callback(slicolmax,maxcol_cb,(long)slicolmin);


  fl_end_form();

  return fdui;
}
/*---------------------------------------*/


void convertMulti2XPM(MultiStruct *My3D)
{
  int ii,kk,index,iu,il,off;
  float *mysignal,scal;
  float minvol,maxvol;
  int XX,YY,ZZ;

  XX=X*zoomfactorx;
  YY=Y*zoomfactory;
  ZZ=Z*zoomfactorz;

  minvol=MinMultiStruct(My3D);
  maxvol=MaxMultiStruct(My3D);

  mysignal=My3D->Signal;

  myvol=(char *)malloc((My3D->NoOfSamples)*2);


  myxyimage=(char **)malloc((1+256+YY)*sizeof(char*));
  myxyimage[0]=(char *)malloc(30);
  sprintf(myxyimage[0],"%i %i 256 2",XX,YY);
  for (ii=0;ii<256;ii++)
    myxyimage[ii+1]=mygraycolormap[ii];
  
  for (ii=0;ii<YY;ii++){
    off = ii+257;
    if ((myxyimage[off]=(char *)malloc(2*XX+1))==NULL) 
      Error("No Malloc");
    myxyimage[off][2*X]='\0';
  }

  myyzimage=(char **)malloc((1+256+ZZ)*sizeof(char*));
  myyzimage[0]=(char *)malloc(30);
  sprintf(myyzimage[0],"%i %i 256 2",YY,ZZ);
  for (ii=0;ii<256;ii++)
    myyzimage[ii+1]=mygraycolormap[ii];
  
  for (ii=0;ii<ZZ;ii++){
    off = ii+257;
    if ((myyzimage[off]=(char *)malloc(2*YY+1))==NULL) 
      Error("No Malloc");
    myyzimage[off][2*Y]='\0';
  }

  myxzimage=(char **)malloc((1+256+ZZ)*sizeof(char*));
  myxzimage[0]=(char *)malloc(30);
  sprintf(myxzimage[0],"%i %i 256 2",XX,ZZ);
  for (ii=0;ii<256;ii++)
    myxzimage[ii+1]=mygraycolormap[ii];
  
  for (ii=0;ii<ZZ;ii++){
    off = ii+257;
    if ((myxzimage[off]=(char *)malloc(2*XX+1))==NULL) 
      Error("No Malloc");
    myxzimage[off][2*X]='\0';
  }

  scal=255.0/(maxvol-minvol);
  kk=0;
  for (ii=0;ii<My3D->NoOfSamples;ii++) 
  {
    index=(int)(scal*(mysignal[ii]-minvol));
    iu=(int)(index/16);
    il=index%16;
    myvol[kk++]=65+iu;
    myvol[kk++]=65+il;
  }
}


int main(int argc, char *argv[])
{
   FD_vol *fd_vol;
   MultiStruct *My3D;
   char Temp[100],*chrptr;
   int filter;

   DebugLevel=_DNormal;
   OpenLog("vol_view");
   if (argc<2)
     Error("Usage vol_view multistruct-filename zoomx zoomy zoomz filter-half-width");
   
   zoomfactorx=zoomfactory=zoomfactorz=1;

   mincol=0;
   maxcol=255;
   mygraycolormap=(char **)malloc(256*sizeof(char*));
   myhotcolormap=(char **)malloc(256*sizeof(char*));
   myrgbcolormap=(char **)malloc(256*sizeof(char*));

   if (argc>4)
   {
     sscanf(argv[2],"%i",&zoomfactorx);
     if ((zoomfactorx<1) || (zoomfactorx>maxzoomfactor))
       Error("Illegal zoomfactorx");
   }

   if (argc>4)
   {
     sscanf(argv[3],"%i",&zoomfactory);
     if ((zoomfactory<1) || (zoomfactory>maxzoomfactor))
       Error("Illegal zoomfactory");
   }
   if (argc>4)
   {
     sscanf(argv[4],"%i",&zoomfactorz);
     if ((zoomfactorz<1) || (zoomfactorz>maxzoomfactor))
       Error("Illegal zoomfactorz");
   }

   strcpy(Temp,argv[1]);
   if((chrptr=strrchr(Temp,'.')))
     *chrptr='\0';
     My3D=ReadMultiStruct(Temp,3);

   if (argc==6)
   {
     sscanf(argv[5],"%i",&filter);
     if (filter>1)
       BoxFilterVolume(My3D,filter);
   }

   X=My3D->Samples[0];
   Y=My3D->Samples[1];
   Z=My3D->Samples[2];
   
   InfoMultiStruct(_DNormal,My3D);

   makegraycolormap();
   makehotcolormap();
   makergbcolormap();

   convertMulti2XPM(My3D);

   FreeMultiStruct(My3D);

   fl_initialize(&argc, argv, 0, 0, 0);
   fd_vol = create_form_vol();
  
   fl_call_object_callback(slix);
   fl_call_object_callback(sliy);
   fl_call_object_callback(sliz);

   /* show the first form */
   fl_show_form(fd_vol->vol,FL_PLACE_CENTER,FL_FULLBORDER,"Vol_view");
   fl_do_forms();
   CloseLog();
   return 0;
}
